#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from oslo_log import versionutils
from oslo_policy import policy

from manila.common import constants
from manila.policies import base


BASE_POLICY_NAME = 'share:%s'

DEPRECATED_REASON = """
The share API now supports scope and default roles.
"""

# Deprecated share policies
deprecated_share_create = policy.DeprecatedRule(
    name=BASE_POLICY_NAME % 'create',
    check_str="",
    deprecated_reason=DEPRECATED_REASON,
    deprecated_since=versionutils.deprecated.WALLABY
)
deprecated_share_create_public = policy.DeprecatedRule(
    name=BASE_POLICY_NAME % 'create_public_share',
    check_str=base.RULE_ADMIN_API,
    deprecated_reason=DEPRECATED_REASON,
    deprecated_since=versionutils.deprecated.WALLABY
)
deprecated_share_get = policy.DeprecatedRule(
    name=BASE_POLICY_NAME % 'get',
    check_str=base.RULE_DEFAULT,
    deprecated_reason=DEPRECATED_REASON,
    deprecated_since=versionutils.deprecated.WALLABY
)
deprecated_share_get_all = policy.DeprecatedRule(
    name=BASE_POLICY_NAME % 'get_all',
    check_str=base.RULE_DEFAULT,
    deprecated_reason=DEPRECATED_REASON,
    deprecated_since=versionutils.deprecated.WALLABY
)
deprecated_share_update = policy.DeprecatedRule(
    name=BASE_POLICY_NAME % 'update',
    check_str=base.RULE_DEFAULT,
    deprecated_reason=DEPRECATED_REASON,
    deprecated_since=versionutils.deprecated.WALLABY
)
deprecated_share_set_public = policy.DeprecatedRule(
    name=BASE_POLICY_NAME % 'set_public_share',
    check_str=base.RULE_ADMIN_API,
    deprecated_reason=DEPRECATED_REASON,
    deprecated_since=versionutils.deprecated.WALLABY
)
deprecated_share_delete = policy.DeprecatedRule(
    name=BASE_POLICY_NAME % 'delete',
    check_str=base.RULE_DEFAULT,
    deprecated_reason=DEPRECATED_REASON,
    deprecated_since=versionutils.deprecated.WALLABY
)
deprecated_share_force_delete = policy.DeprecatedRule(
    name=BASE_POLICY_NAME % 'force_delete',
    check_str=base.RULE_ADMIN_API,
    deprecated_reason=DEPRECATED_REASON,
    deprecated_since=versionutils.deprecated.WALLABY
)
deprecated_share_manage = policy.DeprecatedRule(
    name=BASE_POLICY_NAME % 'manage',
    check_str=base.RULE_ADMIN_API,
    deprecated_reason=DEPRECATED_REASON,
    deprecated_since=versionutils.deprecated.WALLABY
)
deprecated_share_unmanage = policy.DeprecatedRule(
    name=BASE_POLICY_NAME % 'unmanage',
    check_str=base.RULE_ADMIN_API,
    deprecated_reason=DEPRECATED_REASON,
    deprecated_since=versionutils.deprecated.WALLABY
)
deprecated_share_list_by_host = policy.DeprecatedRule(
    name=BASE_POLICY_NAME % 'list_by_host',
    check_str=base.RULE_ADMIN_API,
    deprecated_reason=DEPRECATED_REASON,
    deprecated_since=versionutils.deprecated.WALLABY
)
deprecated_share_list_by_server_id = policy.DeprecatedRule(
    name=BASE_POLICY_NAME % 'list_by_share_server_id',
    check_str=base.RULE_ADMIN_API,
    deprecated_reason=DEPRECATED_REASON,
    deprecated_since=versionutils.deprecated.WALLABY
)
deprecated_share_access_get = policy.DeprecatedRule(
    name=BASE_POLICY_NAME % 'access_get',
    check_str=base.RULE_DEFAULT,
    deprecated_reason=DEPRECATED_REASON,
    deprecated_since=versionutils.deprecated.WALLABY
)
deprecated_share_access_get_all = policy.DeprecatedRule(
    name=BASE_POLICY_NAME % 'access_get_all',
    check_str=base.RULE_DEFAULT,
    deprecated_reason=DEPRECATED_REASON,
    deprecated_since=versionutils.deprecated.WALLABY
)
deprecated_share_extend = policy.DeprecatedRule(
    name=BASE_POLICY_NAME % 'extend',
    check_str=base.RULE_DEFAULT,
    deprecated_reason=DEPRECATED_REASON,
    deprecated_since=versionutils.deprecated.WALLABY
)
deprecated_share_shrink = policy.DeprecatedRule(
    name=BASE_POLICY_NAME % 'shrink',
    check_str=base.RULE_DEFAULT,
    deprecated_reason=DEPRECATED_REASON,
    deprecated_since=versionutils.deprecated.WALLABY
)
deprecated_share_migration_start = policy.DeprecatedRule(
    name=BASE_POLICY_NAME % 'migration_start',
    check_str=base.RULE_ADMIN_API,
    deprecated_reason=DEPRECATED_REASON,
    deprecated_since=versionutils.deprecated.WALLABY
)
deprecated_share_migration_complete = policy.DeprecatedRule(
    name=BASE_POLICY_NAME % 'migration_complete',
    check_str=base.RULE_ADMIN_API,
    deprecated_reason=DEPRECATED_REASON,
    deprecated_since=versionutils.deprecated.WALLABY
)
deprecated_share_migration_cancel = policy.DeprecatedRule(
    name=BASE_POLICY_NAME % 'migration_cancel',
    check_str=base.RULE_ADMIN_API,
    deprecated_reason=DEPRECATED_REASON,
    deprecated_since=versionutils.deprecated.WALLABY
)
deprecated_share_migration_get_progress = policy.DeprecatedRule(
    name=BASE_POLICY_NAME % 'migration_get_progress',
    check_str=base.RULE_ADMIN_API,
    deprecated_reason=DEPRECATED_REASON,
    deprecated_since=versionutils.deprecated.WALLABY
)
deprecated_share_reset_task_state = policy.DeprecatedRule(
    name=BASE_POLICY_NAME % 'reset_task_state',
    check_str=base.RULE_ADMIN_API,
    deprecated_reason=DEPRECATED_REASON,
    deprecated_since=versionutils.deprecated.WALLABY
)
deprecated_share_reset_status = policy.DeprecatedRule(
    name=BASE_POLICY_NAME % 'reset_status',
    check_str=base.RULE_ADMIN_API,
    deprecated_reason=DEPRECATED_REASON,
    deprecated_since=versionutils.deprecated.WALLABY
)
deprecated_share_revert_to_snapshot = policy.DeprecatedRule(
    name=BASE_POLICY_NAME % 'revert_to_snapshot',
    check_str=base.RULE_DEFAULT,
    deprecated_reason=DEPRECATED_REASON,
    deprecated_since=versionutils.deprecated.WALLABY
)
deprecated_share_allow_access = policy.DeprecatedRule(
    name=BASE_POLICY_NAME % 'allow_access',
    check_str=base.RULE_DEFAULT,
    deprecated_reason=DEPRECATED_REASON,
    deprecated_since=versionutils.deprecated.WALLABY
)
deprecated_share_deny_access = policy.DeprecatedRule(
    name=BASE_POLICY_NAME % 'deny_access',
    check_str=base.RULE_DEFAULT,
    deprecated_reason=DEPRECATED_REASON,
    deprecated_since=versionutils.deprecated.WALLABY
)
deprecated_share_update_metadata = policy.DeprecatedRule(
    name=BASE_POLICY_NAME % 'update_share_metadata',
    check_str=base.RULE_DEFAULT,
    deprecated_reason=DEPRECATED_REASON,
    deprecated_since=versionutils.deprecated.WALLABY
)
deprecated_share_delete_metadata = policy.DeprecatedRule(
    name=BASE_POLICY_NAME % 'delete_share_metadata',
    check_str=base.RULE_DEFAULT,
    deprecated_reason=DEPRECATED_REASON,
    deprecated_since=versionutils.deprecated.WALLABY
)
deprecated_share_get_metadata = policy.DeprecatedRule(
    name=BASE_POLICY_NAME % 'get_share_metadata',
    check_str=base.RULE_DEFAULT,
    deprecated_reason=DEPRECATED_REASON,
    deprecated_since=versionutils.deprecated.WALLABY
)

# deprecated legacy snapshot policies with "share" as base resource
deprecated_share_create_snapshot = policy.DeprecatedRule(
    name=BASE_POLICY_NAME % 'create_snapshot',
    check_str=base.RULE_DEFAULT,
    deprecated_reason=DEPRECATED_REASON,
    deprecated_since=versionutils.deprecated.WALLABY
)
deprecated_share_delete_snapshot = policy.DeprecatedRule(
    name=BASE_POLICY_NAME % 'delete_snapshot',
    check_str=base.RULE_DEFAULT,
    deprecated_reason=DEPRECATED_REASON,
    deprecated_since=versionutils.deprecated.WALLABY
)
deprecated_share_snapshot_update = policy.DeprecatedRule(
    name=BASE_POLICY_NAME % 'snapshot_update',
    check_str=base.RULE_DEFAULT,
    deprecated_reason=DEPRECATED_REASON,
    deprecated_since=versionutils.deprecated.WALLABY
)

deprecated_update_admin_only_metadata = policy.DeprecatedRule(
    name=BASE_POLICY_NAME % 'update_admin_only_metadata',
    check_str=base.RULE_ADMIN_API,
    deprecated_reason=DEPRECATED_REASON,
    deprecated_since="YOGA"
)

deprecated_list_shares_in_deferred_deletion_states = policy.DeprecatedRule(
    name=BASE_POLICY_NAME % 'list_shares_in_deferred_deletion_states',
    check_str=base.RULE_ADMIN_API,
    deprecated_reason=DEPRECATED_REASON,
    deprecated_since='2024.1/Caracal'
)

shares_policies = [
    policy.DocumentedRuleDefault(
        name=BASE_POLICY_NAME % 'create',
        check_str=base.ADMIN_OR_PROJECT_MEMBER,
        scope_types=['project'],
        description="Create share.",
        operations=[
            {
                'method': 'POST',
                'path': '/shares',
            }
        ],
        deprecated_rule=deprecated_share_create
    ),
    policy.DocumentedRuleDefault(
        name=BASE_POLICY_NAME % 'create_public_share',
        check_str=base.ADMIN,
        scope_types=['project'],
        description="Create shares visible across all projects in the cloud.",
        operations=[
            {
                'method': 'POST',
                'path': '/shares',
            }
        ],
        deprecated_rule=deprecated_share_create_public
    ),
    policy.DocumentedRuleDefault(
        name=BASE_POLICY_NAME % 'get',
        check_str=base.ADMIN_OR_PROJECT_READER,
        scope_types=['project'],
        description="Get share.",
        operations=[
            {
                'method': 'GET',
                'path': '/shares/{share_id}',
            }
        ],
        deprecated_rule=deprecated_share_get
    ),
    policy.DocumentedRuleDefault(
        name=BASE_POLICY_NAME % 'get_all',
        check_str=base.ADMIN_OR_PROJECT_READER,
        scope_types=['project'],
        description="List shares.",
        operations=[
            {
                'method': 'GET',
                'path': '/shares?{query}',
            },
            {
                'method': 'GET',
                'path': '/shares/detail?{query}',
            }
        ],
        deprecated_rule=deprecated_share_get_all
    ),
    policy.DocumentedRuleDefault(
        name=BASE_POLICY_NAME % 'update',
        check_str=base.ADMIN_OR_PROJECT_MEMBER,
        scope_types=['project'],
        description="Update a share.",
        operations=[
            {
                'method': 'PUT',
                'path': '/shares/{share_id}',
            }
        ],
        deprecated_rule=deprecated_share_update
    ),
    policy.DocumentedRuleDefault(
        name=BASE_POLICY_NAME % 'set_public_share',
        check_str=base.ADMIN,
        scope_types=['project'],
        description="Update a share to be visible across all projects in the "
                    "cloud.",
        operations=[
            {
                'method': 'PUT',
                'path': '/shares/{share_id}',
            }
        ],
        deprecated_rule=deprecated_share_set_public
    ),
    policy.DocumentedRuleDefault(
        name=BASE_POLICY_NAME % 'delete',
        check_str=base.ADMIN_OR_PROJECT_MEMBER,
        scope_types=['project'],
        description="Delete share.",
        operations=[
            {
                'method': 'DELETE',
                'path': '/shares/{share_id}',
            }
        ],
        deprecated_rule=deprecated_share_delete
    ),
    policy.DocumentedRuleDefault(
        name=BASE_POLICY_NAME % 'soft_delete',
        check_str=base.ADMIN_OR_PROJECT_MEMBER,
        scope_types=['project'],
        description="Soft Delete a share.",
        operations=[
            {
                'method': 'POST',
                'path': '/shares/{share_id}/action',
            }
        ],
    ),
    policy.DocumentedRuleDefault(
        name=BASE_POLICY_NAME % 'restore',
        check_str=base.ADMIN_OR_PROJECT_MEMBER,
        scope_types=['project'],
        description="Restore a share.",
        operations=[
            {
                'method': 'POST',
                'path': '/shares/{share_id}/action',
            }
        ],
    ),
    policy.DocumentedRuleDefault(
        name=BASE_POLICY_NAME % 'force_delete',
        check_str=base.ADMIN,
        scope_types=['project'],
        description="Force Delete a share.",
        operations=[
            {
                'method': 'DELETE',
                'path': '/shares/{share_id}',
            }
        ],
        deprecated_rule=deprecated_share_force_delete
    ),
    policy.DocumentedRuleDefault(
        name=BASE_POLICY_NAME % 'manage',
        check_str=base.ADMIN,
        scope_types=['project'],
        description="Manage share.",
        operations=[
            {
                'method': 'POST',
                'path': '/shares/manage',
            }
        ],
        deprecated_rule=deprecated_share_manage
    ),
    policy.DocumentedRuleDefault(
        name=BASE_POLICY_NAME % 'unmanage',
        check_str=base.ADMIN,
        scope_types=['project'],
        description="Unmanage share.",
        operations=[
            {
                'method': 'POST',
                'path': '/shares/unmanage',
            }
        ],
        deprecated_rule=deprecated_share_unmanage
    ),
    policy.DocumentedRuleDefault(
        name=BASE_POLICY_NAME % 'list_by_host',
        check_str=base.ADMIN,
        scope_types=['project'],
        description="List share by host.",
        operations=[
            {
                'method': 'GET',
                'path': '/shares?host={host}',
            },
            {
                'method': 'GET',
                'path': '/shares/detail?host={host}',
            }
        ],
        deprecated_rule=deprecated_share_list_by_host
    ),
    policy.DocumentedRuleDefault(
        name=BASE_POLICY_NAME % 'list_by_share_server_id',
        check_str=base.ADMIN,
        scope_types=['project'],
        description="List share by server id.",
        operations=[
            {
                'method': 'GET',
                'path': '/shares?share_server_id={share_server_id}'
            },
            {
                'method': 'GET',
                'path': '/shares/detail?share_server_id={share_server_id}',
            }
        ],
        deprecated_rule=deprecated_share_list_by_server_id
    ),
    policy.DocumentedRuleDefault(
        name=BASE_POLICY_NAME % 'access_get',
        check_str=base.ADMIN_OR_PROJECT_READER,
        scope_types=['project'],
        description="Get share access rule (deprecated in API version 2.45).",
        operations=[
            {
                'method': 'POST',
                'path': '/shares/{share_id}/action',
            }
        ],
        deprecated_rule=deprecated_share_access_get
    ),
    policy.DocumentedRuleDefault(
        name=BASE_POLICY_NAME % 'access_get_all',
        check_str=base.ADMIN_OR_PROJECT_READER,
        scope_types=['project'],
        description=("List share access rules (deprecated in API "
                     "version 2.45)."),
        operations=[
            {
                'method': 'GET',
                'path': '/shares/{share_id}/action',
            }
        ],
        deprecated_rule=deprecated_share_access_get_all
    ),
    policy.DocumentedRuleDefault(
        name=BASE_POLICY_NAME % 'extend',
        check_str=base.ADMIN_OR_PROJECT_MEMBER,
        scope_types=['project'],
        description="Extend share.",
        operations=[
            {
                'method': 'POST',
                'path': '/shares/{share_id}/action',
            }
        ],
        deprecated_rule=deprecated_share_extend
    ),
    policy.DocumentedRuleDefault(
        name=BASE_POLICY_NAME % 'force_extend',
        check_str=base.ADMIN,
        scope_types=['project'],
        description="Force extend share.",
        operations=[
            {
                'method': 'POST',
                'path': '/shares/{share_id}/action',
            }
        ]),
    policy.DocumentedRuleDefault(
        name=BASE_POLICY_NAME % constants.POLICY_EXTEND_BEYOND_MAX_SHARE_SIZE,
        check_str=base.ADMIN,
        scope_types=['project'],
        description="Extend share beyond max share size.",
        operations=[
            {
                'method': 'POST',
                'path': '/shares/{share_id}/action',
            }
        ]),
    policy.DocumentedRuleDefault(
        name=BASE_POLICY_NAME % 'shrink',
        check_str=base.ADMIN_OR_PROJECT_MEMBER,
        scope_types=['project'],
        description="Shrink share.",
        operations=[
            {
                'method': 'POST',
                'path': '/shares/{share_id}/action',
            }
        ],
        deprecated_rule=deprecated_share_shrink
    ),
    policy.DocumentedRuleDefault(
        name=BASE_POLICY_NAME % 'migration_start',
        check_str=base.ADMIN,
        scope_types=['project'],
        description="Migrate a share to the specified host.",
        operations=[
            {
                'method': 'POST',
                'path': '/shares/{share_id}/action',
            }
        ],
        deprecated_rule=deprecated_share_migration_start
    ),
    policy.DocumentedRuleDefault(
        name=BASE_POLICY_NAME % 'migration_complete',
        check_str=base.ADMIN,
        scope_types=['project'],
        description="Invoke 2nd phase of share migration.",
        operations=[
            {
                'method': 'POST',
                'path': '/shares/{share_id}/action',
            }
        ],
        deprecated_rule=deprecated_share_migration_complete
    ),
    policy.DocumentedRuleDefault(
        name=BASE_POLICY_NAME % 'migration_cancel',
        check_str=base.ADMIN,
        scope_types=['project'],
        description="Attempt to cancel share migration.",
        operations=[
            {
                'method': 'POST',
                'path': '/shares/{share_id}/action',
            }
        ],
        deprecated_rule=deprecated_share_migration_cancel
    ),
    policy.DocumentedRuleDefault(
        name=BASE_POLICY_NAME % 'migration_get_progress',
        check_str=base.ADMIN,
        scope_types=['project'],
        description=("Retrieve share migration progress for a given "
                     "share."),
        operations=[
            {
                'method': 'POST',
                'path': '/shares/{share_id}/action',
            }
        ],
        deprecated_rule=deprecated_share_migration_get_progress
    ),
    policy.DocumentedRuleDefault(
        name=BASE_POLICY_NAME % 'reset_task_state',
        check_str=base.ADMIN,
        scope_types=['project'],
        description="Reset task state.",
        operations=[
            {
                'method': 'POST',
                'path': '/shares/{share_id}/action',
            }
        ],
        deprecated_rule=deprecated_share_reset_task_state
    ),
    policy.DocumentedRuleDefault(
        name=BASE_POLICY_NAME % 'reset_status',
        check_str=base.ADMIN,
        scope_types=['project'],
        description="Reset status.",
        operations=[
            {
                'method': 'POST',
                'path': '/shares/{share_id}/action',
            }
        ],
        deprecated_rule=deprecated_share_reset_status
    ),
    policy.DocumentedRuleDefault(
        name=BASE_POLICY_NAME % 'revert_to_snapshot',
        check_str=base.ADMIN_OR_PROJECT_MEMBER,
        scope_types=['project'],
        description="Revert a share to a snapshot.",
        operations=[
            {
                'method': 'POST',
                'path': '/shares/{share_id}/action',
            }
        ],
        deprecated_rule=deprecated_share_revert_to_snapshot
    ),
    policy.DocumentedRuleDefault(
        name=BASE_POLICY_NAME % 'allow_access',
        check_str=base.ADMIN_OR_PROJECT_MEMBER,
        scope_types=['project'],
        description="Add share access rule.",
        operations=[
            {
                'method': 'POST',
                'path': '/shares/{share_id}/action',
            }
        ],
        deprecated_rule=deprecated_share_allow_access
    ),
    policy.DocumentedRuleDefault(
        name=BASE_POLICY_NAME % 'deny_access',
        check_str=base.ADMIN_OR_PROJECT_MEMBER,
        scope_types=['project'],
        description="Remove share access rule.",
        operations=[
            {
                'method': 'POST',
                'path': '/shares/{share_id}/action',
            }
        ],
        deprecated_rule=deprecated_share_deny_access
    ),
    policy.DocumentedRuleDefault(
        name=BASE_POLICY_NAME % 'update_share_metadata',
        check_str=base.ADMIN_OR_PROJECT_MEMBER,
        scope_types=['project'],
        description="Update share metadata.",
        operations=[
            {
                'method': 'PUT',
                'path': '/shares/{share_id}/metadata',
            },
            {
                'method': 'POST',
                'path': '/shares/{share_id}/metadata/{key}',
            },
            {
                'method': 'POST',
                'path': '/shares/{share_id}/metadata',
            },
        ],
        deprecated_rule=deprecated_share_update_metadata
    ),
    policy.DocumentedRuleDefault(
        name=BASE_POLICY_NAME % 'delete_share_metadata',
        check_str=base.ADMIN_OR_PROJECT_MEMBER,
        scope_types=['project'],
        description="Delete share metadata.",
        operations=[
            {
                'method': 'DELETE',
                'path': '/shares/{share_id}/metadata/{key}',
            }
        ],
        deprecated_rule=deprecated_share_delete_metadata
    ),
    policy.DocumentedRuleDefault(
        name=BASE_POLICY_NAME % 'get_share_metadata',
        check_str=base.ADMIN_OR_PROJECT_READER,
        scope_types=['project'],
        description="Get share metadata.",
        operations=[
            {
                'method': 'GET',
                'path': '/shares/{share_id}/metadata',
            },
            {
                'method': 'GET',
                'path': '/shares/{share_id}/metadata/{key}',
            }
        ],
        deprecated_rule=deprecated_share_get_metadata
    ),
    policy.DocumentedRuleDefault(
        name=BASE_POLICY_NAME % 'list_shares_in_deferred_deletion_states',
        check_str=base.ADMIN,
        scope_types=['project'],
        description="List shares whose deletion has been deferred",
        operations=[
            {
                'method': 'GET',
                'path': '/v2/shares',
            },
        ],
        deprecated_rule=deprecated_list_shares_in_deferred_deletion_states
    ),

]

# NOTE(gouthamr) For historic reasons, some snapshot policies used
# "share" as the resource. We could deprecate these and move them to using
# "share_snapshot" as the base resource in the future.
base_snapshot_policies = [
    policy.DocumentedRuleDefault(
        name=BASE_POLICY_NAME % 'create_snapshot',
        check_str=base.ADMIN_OR_PROJECT_MEMBER,
        scope_types=['project'],
        description="Create share snapshot.",
        operations=[
            {
                'method': 'POST',
                'path': '/snapshots',
            }
        ],
        deprecated_rule=deprecated_share_create_snapshot
    ),
    policy.DocumentedRuleDefault(
        name=BASE_POLICY_NAME % 'delete_snapshot',
        check_str=base.ADMIN_OR_PROJECT_MEMBER,
        scope_types=['project'],
        description="Delete share snapshot.",
        operations=[
            {
                'method': 'DELETE',
                'path': '/snapshots/{snapshot_id}',
            }
        ],
        deprecated_rule=deprecated_share_delete_snapshot
    ),
    policy.DocumentedRuleDefault(
        name=BASE_POLICY_NAME % 'snapshot_update',
        check_str=base.ADMIN_OR_PROJECT_MEMBER,
        scope_types=['project'],
        description="Update share snapshot.",
        operations=[
            {
                'method': 'PUT',
                'path': '/snapshots/{snapshot_id}/action',
            }
        ],
        deprecated_rule=deprecated_share_snapshot_update
    ),
    policy.DocumentedRuleDefault(
        name=BASE_POLICY_NAME % 'update_admin_only_metadata',
        check_str=base.ADMIN,
        scope_types=['project'],
        description=(
            "Update metadata items that are considered \"admin only\" "
            "by the service."),
        operations=[
            {
                'method': 'PUT',
                'path': '/shares/{share_id}/metadata',
            }
        ],
        deprecated_rule=deprecated_update_admin_only_metadata
    ),
]


def list_rules():
    return shares_policies + base_snapshot_policies



# Copyright 2023 Cloudification GmbH.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from manila.api import common
from manila import policy


class BackupViewBuilder(common.ViewBuilder):
    """Model a server API response as a python dictionary."""

    _collection_name = 'share_backups'
    _collection_links = 'share_backup_links'

    _detail_version_modifiers = [
        "add_backup_type_field",
    ]

    def summary_list(self, request, backups):
        """Summary view of a list of backups."""
        return self._list_view(self.summary, request, backups)

    def detail_list(self, request, backups):
        """Detailed view of a list of backups."""
        return self._list_view(self.detail, request, backups)

    def summary(self, request, backup):
        """Generic, non-detailed view of a share backup."""

        backup_dict = {
            'id': backup.get('id'),
            'name': backup.get('display_name'),
            'share_id': backup.get('share_id'),
            'status': backup.get('status'),
        }
        return {'share_backup': backup_dict}

    def restore_summary(self, request, restore):
        """Generic, non-detailed view of a restore."""
        return {
            'restore': {
                'backup_id': restore['backup_id'],
                'share_id': restore['share_id'],
            },
        }

    def detail(self, request, backup):
        """Detailed view of a single backup."""
        context = request.environ['manila.context']
        backup_dict = {
            'id': backup.get('id'),
            'name': backup.get('display_name'),
            'share_id': backup.get('share_id'),
            'status': backup.get('status'),
            'description': backup.get('display_description'),
            'size': backup.get('size'),
            'created_at': backup.get('created_at'),
            'updated_at': backup.get('updated_at'),
            'availability_zone': backup.get('availability_zone'),
            'progress': backup.get('progress'),
            'restore_progress': backup.get('restore_progress'),
        }

        self.update_versioned_resource_dict(request, backup_dict, backup)
        if policy.check_is_host_admin(context):
            backup_dict['host'] = backup.get('host')
            backup_dict['topic'] = backup.get('topic')

        return {'share_backup': backup_dict}

    def _list_view(self, func, request, backups):
        """Provide a view for a list of backups."""

        backups_list = [func(request, backup)['share_backup']
                        for backup in backups]

        backup_links = self._get_collection_links(
            request, backups, self._collection_name)
        backups_dict = {self._collection_name: backups_list}

        if backup_links:
            backups_dict[self._collection_links] = backup_links

        return backups_dict

    @common.ViewBuilder.versioned_method("2.85")
    def add_backup_type_field(self, context, backup_dict, backup):
        backup_dict['backup_type'] = backup.get('backup_type')



# Copyright (c) 2023 Dell Inc. or its subsidiaries.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from http import client as http_client
import json
import pathlib

import ddt
import requests_mock

from manila import exception
from manila.share.drivers.dell_emc.plugins.powerflex import (
    object_manager as manager
)
from manila import test


@ddt.ddt
class StorageObjectManagerTestCase(test.TestCase):
    def setUp(self):
        super(StorageObjectManagerTestCase, self).setUp()

        self._mock_url = "https://192.168.0.110:443"
        self.manager = manager.StorageObjectManager(
            self._mock_url, username="admin", password="pwd", export_path=None
        )
        self.mockup_file_base = (
            str(pathlib.Path.cwd())
            + "/manila/tests/share/drivers/dell_emc/plugins/powerflex/mockup/"
        )

    @ddt.data(False, True)
    def test__get_headers(self, got_token):
        self.manager.got_token = got_token
        self.manager.rest_token = "token_str"
        self.assertEqual(
            self.manager._get_headers().get("Authorization") is not None,
            got_token,
        )

    def _getJsonFile(self, filename):
        f = open(self.mockup_file_base + filename)
        data = json.load(f)
        f.close()
        return data

    @requests_mock.mock()
    def test_get_nas_server_id(self, m):
        nas_server = "env8nasserver"
        self.assertEqual(0, len(m.request_history))
        self._add_get_nas_server_id_response(
            m, nas_server, self._getJsonFile("get_nas_server_id_response.json")
        )
        id = self.manager.get_nas_server_id(nas_server)
        self.assertEqual(id, "64132f37-d33e-9d4a-89ba-d625520a4779")

    def _add_get_nas_server_id_response(self, m, nas_server, json_str):
        url = "{0}/rest/v1/nas-servers?select=id&name=eq.{1}".format(
            self._mock_url, nas_server
        )
        m.get(url, status_code=200, json=json_str)

    @requests_mock.mock()
    def test_create_filesystem(self, m):
        nas_server = "env8nasserver"
        self.assertEqual(0, len(m.request_history))
        self._add_get_nas_server_id_response(
            m, nas_server, self._getJsonFile("get_nas_server_id_response.json")
        )
        storage_pool_id = "8515fee00000000"
        self._add_create_filesystem_response(
            m, self._getJsonFile("create_filesystem_response.json")
        )
        id = self.manager.create_filesystem(
            storage_pool_id,
            nas_server,
            name="Manila-filesystem",
            size=3221225472,
        )
        self.assertEqual(id, "6432b79e-1cc3-0414-3ffd-2a50fb1ccff3")

    def _add_create_filesystem_response(self, m, json_str):
        url = "{0}/rest/v1/file-systems".format(self._mock_url)
        m.post(url, status_code=201, json=json_str)

    @requests_mock.mock()
    def test_create_nfs_export(self, m):
        filesystem_id = "6432b79e-1cc3-0414-3ffd-2a50fb1ccff3"
        name = "Manila-UT-filesystem"
        self.assertEqual(0, len(m.request_history))
        self._add_create_nfs_export_response(
            m, self._getJsonFile("create_nfs_export_response.json")
        )
        id = self.manager.create_nfs_export(filesystem_id, name)
        self.assertEqual(id, "6433a2b2-6d60-f737-9f3b-2a50fb1ccff3")

    def _add_create_nfs_export_response(self, m, json_str):
        url = "{0}/rest/v1/nfs-exports".format(self._mock_url)
        m.post(url, status_code=201, json=json_str)

    @requests_mock.mock()
    def test_delete_filesystem(self, m):
        filesystem_id = "6432b79e-1cc3-0414-3ffd-2a50fb1ccff3"
        self.assertEqual(0, len(m.request_history))
        self._add_delete_filesystem_response(m, filesystem_id)
        result = self.manager.delete_filesystem(filesystem_id)
        self.assertEqual(result, True)

    def _add_delete_filesystem_response(self, m, filesystem_id):
        url = "{0}/rest/v1/file-systems/{1}".format(
            self._mock_url, filesystem_id
        )
        m.delete(url, status_code=204)

    @requests_mock.mock()
    def test_create_snapshot(self, m):
        name = "Manila-UT-filesystem-snap"
        filesystem_id = "6432b79e-1cc3-0414-3ffd-2a50fb1ccff3"
        self.assertEqual(0, len(m.request_history))
        self._add_create_snapshot_response(
            m,
            filesystem_id,
            self._getJsonFile("create_nfs_snapshot_response.json"),
        )
        result = self.manager.create_snapshot(name, filesystem_id)
        self.assertEqual(result, True)

    def _add_create_snapshot_response(self, m, filesystem_id, json_str):
        url = "{0}/rest/v1/file-systems/{1}/snapshot".format(
            self._mock_url, filesystem_id
        )
        m.post(url, status_code=201, json=json_str)

    @requests_mock.mock()
    def test_get_nfs_export_name(self, m):
        export_id = "6433a2b2-6d60-f737-9f3b-2a50fb1ccff3"
        self.assertEqual(0, len(m.request_history))
        self._add_get_nfs_export_name_response(
            m,
            export_id,
            self._getJsonFile("get_nfs_export_name_response.json"),
        )
        name = self.manager.get_nfs_export_name(export_id)
        self.assertEqual(name, "Manila-UT-filesystem")

    def _add_get_nfs_export_name_response(self, m, export_id, json_str):
        url = "{0}/rest/v1/nfs-exports/{1}?select=*".format(
            self._mock_url, export_id
        )
        m.get(url, status_code=200, json=json_str)

    @requests_mock.mock()
    def test_get_filesystem_id(self, m):
        name = "Manila-UT-filesystem"
        self.assertEqual(0, len(m.request_history))
        self._add_get_filesystem_id_response(
            m, name, self._getJsonFile("get_fileystem_id_response.json")
        )
        id = self.manager.get_filesystem_id(name)
        self.assertEqual(id, "6432b79e-1cc3-0414-3ffd-2a50fb1ccff3")

    def _add_get_filesystem_id_response(self, m, name, json_str):
        url = "{0}/rest/v1/file-systems?select=id&name=eq.{1}".format(
            self._mock_url, name
        )
        m.get(url, status_code=200, json=json_str)

    @requests_mock.mock()
    def test_get_nfs_export_id(self, m):
        name = "Manila-UT-filesystem"
        self.assertEqual(0, len(m.request_history))
        self._add_get_nfs_export_id_response(
            m, name, self._getJsonFile("get_nfs_export_id_response.json")
        )
        id = self.manager.get_nfs_export_id(name)
        self.assertEqual(id, "6433a2b2-6d60-f737-9f3b-2a50fb1ccff3")

    def _add_get_nfs_export_id_response(self, m, name, json_str):
        url = "{0}/rest/v1/nfs-exports?select=id&name=eq.{1}".format(
            self._mock_url, name
        )
        m.get(url, status_code=200, json=json_str)

    @requests_mock.mock()
    def test_get_storage_pool_id(self, m):
        protection_domain_name = "Env8-PD-1"
        storage_pool_name = "Env8-SP-SW_SSD-1"
        self.assertEqual(0, len(m.request_history))
        self._add_get_storage_pool_id_response(
            m, self._getJsonFile("get_storage_pool_id_response.json")
        )
        id = self.manager.get_storage_pool_id(
            protection_domain_name, storage_pool_name
        )
        self.assertEqual(id, "28515fee00000000")

    def _add_get_storage_pool_id_response(self, m, json_str):
        url = "{0}/api/types/StoragePool/instances/action/queryIdByKey".format(
            self._mock_url
        )
        m.post(url, status_code=200, json=json_str)

    @requests_mock.mock()
    def test_set_export_access(self, m):
        export_id = "6433a2b2-6d60-f737-9f3b-2a50fb1ccff3"
        rw_hosts = "192.168.1.110"
        ro_hosts = "192.168.1.111"
        self.assertEqual(0, len(m.request_history))
        self._add_set_export_access_response(m, export_id)
        result = self.manager.set_export_access(export_id, rw_hosts, ro_hosts)
        self.assertEqual(result, True)

    def _add_set_export_access_response(self, m, export_id):
        url = "{0}/rest/v1/nfs-exports/{1}".format(self._mock_url, export_id)
        m.patch(url, status_code=204)

    @requests_mock.mock()
    def test_extend_export(self, m):
        filesystem_id = "6432b79e-1cc3-0414-3ffd-2a50fb1ccff3"
        new_size = 6441225472
        self.assertEqual(0, len(m.request_history))
        self._add_extend_export_response(m, filesystem_id)
        result = self.manager.extend_export(filesystem_id, new_size)
        self.assertEqual(result, True)

    def _add_extend_export_response(self, m, filesystem_id):
        url = "{0}/rest/v1/file-systems/{1}".format(
            self._mock_url, filesystem_id
        )
        m.patch(url, status_code=204)

    @requests_mock.mock()
    def test_get_fsid_from_export_name(self, m):
        name = "Manila-UT-filesystem"
        self.assertEqual(0, len(m.request_history))
        self._add_get_fsid_from_export_name_response(
            m,
            name,
            self._getJsonFile("get_fsid_from_export_name_response.json"),
        )
        id = self.manager.get_fsid_from_export_name(name)
        self.assertEqual(id, "6432b79e-1cc3-0414-3ffd-2a50fb1ccff3")

    def _add_get_fsid_from_export_name_response(self, m, name, json_str):
        url = (
            "{0}/rest/v1/nfs-exports?select=file_system_id&name=eq.{1}".format(
                self._mock_url, name
            )
        )
        m.get(url, status_code=200, json=json_str)

    @requests_mock.mock()
    def test_get_fsid_from_snapshot_name(self, m):
        snapshot_name = "Manila-UT-filesystem-snap"
        self.assertEqual(0, len(m.request_history))
        self._add_get_fsid_from_snapshot_name_response(
            m,
            snapshot_name,
            self._getJsonFile("get_fsid_from_snapshot_name_response.json"),
        )
        id = self.manager.get_fsid_from_snapshot_name(snapshot_name)
        self.assertEqual(id, "6433b635-6c1f-878e-6467-2a50fb1ccff3")

    def _add_get_fsid_from_snapshot_name_response(
        self, m, snapshot_name, json_str
    ):
        url = "{0}/rest/v1/file-systems?select=id&name=eq.{1}".format(
            self._mock_url, snapshot_name
        )
        m.get(url, status_code=200, json=json_str)

    @requests_mock.mock()
    def test_check_response_with_login_get(self, m):
        nas_server = "env8nasserver"
        self.assertEqual(0, len(m.request_history))
        self._add_get_nas_server_id_response_list(m, nas_server)
        self._add_login_success_response(m)
        id = self.manager.get_nas_server_id(nas_server)
        self.assertEqual(id, "64132f37-d33e-9d4a-89ba-d625520a4779")

    def _add_get_nas_server_id_response_list(self, m, nas_server):
        url = "{0}/rest/v1/nas-servers?select=id&name=eq.{1}".format(
            self._mock_url, nas_server
        )
        m.get(
            url,
            [
                {"status_code": http_client.UNAUTHORIZED},
                {
                    "status_code": 200,
                    "json": self._getJsonFile(
                        "get_nas_server_id_response.json"
                    ),
                },
            ],
        )

    def _add_login_success_response(self, m):
        url = "{0}/rest/auth/login".format(self._mock_url)
        m.post(
            url, status_code=200, json=self._getJsonFile("login_response.json")
        )

    @requests_mock.mock()
    def test_check_response_with_login_post(self, m):
        filesystem_id = "6432b79e-1cc3-0414-3ffd-2a50fb1ccff3"
        name = "Manila-UT-filesystem"
        self.assertEqual(0, len(m.request_history))
        self._add_create_nfs_export_response_list(m)
        self._add_login_success_response(m)
        id = self.manager.create_nfs_export(filesystem_id, name)
        self.assertEqual(id, "6433a2b2-6d60-f737-9f3b-2a50fb1ccff3")

    def _add_create_nfs_export_response_list(self, m):
        url = "{0}/rest/v1/nfs-exports".format(self._mock_url)
        m.post(
            url,
            [
                {"status_code": http_client.UNAUTHORIZED},
                {
                    "status_code": 201,
                    "json": self._getJsonFile(
                        "create_nfs_export_response.json"
                    ),
                },
            ],
        )

    @requests_mock.mock()
    def test_check_response_with_login_delete(self, m):
        filesystem_id = "6432b79e-1cc3-0414-3ffd-2a50fb1ccff3"
        self.assertEqual(0, len(m.request_history))
        self._add_delete_filesystem_response_list(m, filesystem_id)
        self._add_login_success_response(m)
        result = self.manager.delete_filesystem(filesystem_id)
        self.assertEqual(result, True)

    def _add_delete_filesystem_response_list(self, m, filesystem_id):
        url = "{0}/rest/v1/file-systems/{1}".format(
            self._mock_url, filesystem_id
        )
        m.delete(
            url,
            [{"status_code": http_client.UNAUTHORIZED}, {"status_code": 204}],
        )

    @requests_mock.mock()
    def test_check_response_with_login_patch(self, m):
        filesystem_id = "6432b79e-1cc3-0414-3ffd-2a50fb1ccff3"
        new_size = 6441225472
        self.assertEqual(0, len(m.request_history))
        self._add_extend_export_response_list(m, filesystem_id)
        self._add_login_success_response(m)
        result = self.manager.extend_export(filesystem_id, new_size)
        self.assertEqual(result, True)

    def _add_extend_export_response_list(self, m, filesystem_id):
        url = "{0}/rest/v1/file-systems/{1}".format(
            self._mock_url, filesystem_id
        )
        m.patch(
            url,
            [{"status_code": http_client.UNAUTHORIZED}, {"status_code": 204}],
        )

    @requests_mock.mock()
    def test_check_response_with_invalid_credential(self, m):
        nas_server = "env8nasserver"
        self.assertEqual(0, len(m.request_history))
        self._add_get_nas_server_id_unauthorized_response(m, nas_server)
        self._add_login_fail_response(m)
        self.assertRaises(
            exception.NotAuthorized, self.manager.get_nas_server_id, nas_server
        )

    def _add_get_nas_server_id_unauthorized_response(self, m, nas_server):
        url = "{0}/rest/v1/nas-servers?select=id&name=eq.{1}".format(
            self._mock_url, nas_server
        )
        m.get(url, status_code=http_client.UNAUTHORIZED)

    def _add_login_fail_response(self, m):
        url = "{0}/rest/auth/login".format(self._mock_url)
        m.post(url, status_code=http_client.UNAUTHORIZED)

    @requests_mock.mock()
    def test_execute_powerflex_post_request_with_no_param(self, m):
        url = self._mock_url + "/fake_url"
        self.assertEqual(0, len(m.request_history))
        m.post(url, status_code=201)
        res, response = self.manager.execute_powerflex_post_request(url)
        self.assertEqual(res.status_code, 201)

    @requests_mock.mock()
    def test_execute_powerflex_patch_request_with_no_param(self, m):
        url = self._mock_url + "/fake_url"
        self.assertEqual(0, len(m.request_history))
        m.patch(url, status_code=204)
        res = self.manager.execute_powerflex_patch_request(url)
        self.assertEqual(res.status_code, 204)

    @requests_mock.mock()
    def test_get_storage_pool_spare_percentage(self, m):
        storage_pool_id = "28515fee00000000"
        self.assertEqual(0, len(m.request_history))
        self._add_get_storage_pool_spare_percentage(
            m,
            storage_pool_id,
            self._getJsonFile("get_storage_pool_spare_percentage.json"),
        )
        spare = self.manager.get_storage_pool_spare_percentage(storage_pool_id)
        self.assertEqual(spare, 34)

    def _add_get_storage_pool_spare_percentage(self, m, storage_pool_id,
                                               json_str):
        url = (
            "{0}/api/instances/StoragePool::{1}".format(
                self._mock_url, storage_pool_id
            )
        )
        m.get(url, status_code=200, json=json_str)

    @requests_mock.mock()
    def test_get_storage_pool_statistic(self, m):
        storage_pool_id = "28515fee00000000"
        self.assertEqual(0, len(m.request_history))
        self._add_get_storage_pool_statistic(
            m,
            storage_pool_id,
            self._getJsonFile("get_storage_pool_statistic.json"),
        )
        statistic = self.manager.get_storage_pool_statistic(storage_pool_id)
        self.assertEqual(statistic['maxCapacityInKb'], 4826330112)
        self.assertEqual(statistic['capacityInUseInKb'], 53217280)
        self.assertEqual(statistic['netUnusedCapacityInKb'], 1566080512)
        self.assertEqual(statistic['primaryVacInKb'], 184549376)

    def _add_get_storage_pool_statistic(self, m, storage_pool_id,
                                        json_str):
        url = (
            ("{0}/api/instances/StoragePool::{1}/relationships/" +
             "Statistics").format(
                self._mock_url, storage_pool_id
            )
        )
        m.get(url, status_code=200, json=json_str)



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

"""add state column for service

Revision ID: aebe2a413e13
Revises: ac0620cbe74d
Create Date: 2023-01-10 11:43:24.741726

"""

# revision identifiers, used by Alembic.
revision = 'aebe2a413e13'
down_revision = 'ac0620cbe74d'

from alembic import op
from oslo_log import log
import sqlalchemy as sa

LOG = log.getLogger(__name__)


def upgrade():
    try:
        op.add_column(
            'services',
            sa.Column('state', sa.String(36), nullable=True))
    except Exception:
        LOG.error("services table column state not added")
        raise


def downgrade():
    try:
        op.drop_column('services', 'state')
    except Exception:
        LOG.error("services table column state not dropped")
        raise



# Copyright (c) 2023 Dell Inc. or its subsidiaries.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""
PowerStore specific NAS backend plugin.
"""
from oslo_config import cfg
from oslo_log import log
from oslo_utils import units

from manila.common import constants as const
from manila import exception
from manila.i18n import _
from manila.share.drivers.dell_emc.plugins import base as driver
from manila.share.drivers.dell_emc.plugins.powerstore import client

"""Version history:
    1.0 - Initial version
"""
VERSION = "1.0"

CONF = cfg.CONF

LOG = log.getLogger(__name__)

POWERSTORE_OPTS = [
    cfg.StrOpt('dell_nas_backend_host',
               help='Dell NAS backend hostname or IP address.'),
    cfg.StrOpt('dell_nas_server',
               help='Root directory or NAS server which owns the shares.'),
    cfg.StrOpt('dell_ad_domain',
               help='Domain name of the active directory '
               'joined by the NAS server.'),
    cfg.StrOpt('dell_nas_login',
               help='User name for the Dell NAS backend.'),
    cfg.StrOpt('dell_nas_password',
               secret=True,
               help='Password for the Dell NAS backend.'),
    cfg.BoolOpt('dell_ssl_cert_verify',
                default=False,
                help='If set to False the https client will not validate the '
                     'SSL certificate of the backend endpoint.'),
    cfg.StrOpt('dell_ssl_cert_path',
               help='Can be used to specify a non default path to a '
                    'CA_BUNDLE file or directory with certificates of trusted '
                    'CAs, which will be used to validate the backend.')
]


class PowerStoreStorageConnection(driver.StorageConnection):
    """Implements PowerStore specific functionality for Dell Manila driver."""

    def __init__(self, *args, **kwargs):
        """Do initialization"""

        LOG.debug('Invoking base constructor for Manila'
                  ' Dell PowerStore Driver.')
        super(PowerStoreStorageConnection,
              self).__init__(*args, **kwargs)

        LOG.debug('Setting up attributes for Manila'
                  ' Dell PowerStore Driver.')
        if 'configuration' in kwargs:
            kwargs['configuration'].append_config_values(POWERSTORE_OPTS)

        self.client = None
        self.verify_certificate = None
        self.certificate_path = None
        self.ipv6_implemented = True
        self.revert_to_snap_support = True
        self.shrink_share_support = True

        # props from super class
        self.driver_handles_share_servers = False
        # props for share status update
        self.reserved_percentage = None
        self.reserved_snapshot_percentage = None
        self.reserved_share_extend_percentage = None
        self.max_over_subscription_ratio = None

    def connect(self, dell_share_driver, context):
        """Connects to Dell PowerStore"""
        LOG.debug('Reading configuration parameters for Manila'
                  ' Dell PowerStore Driver.')
        config = dell_share_driver.configuration
        get_config_value = config.safe_get
        self.rest_ip = get_config_value("dell_nas_backend_host")
        self.rest_username = get_config_value("dell_nas_login")
        self.rest_password = get_config_value("dell_nas_password")
        # validate IP, username and password
        if not all([self.rest_ip,
                    self.rest_username,
                    self.rest_password]):
            message = _("REST server IP, username and password"
                        " must be specified.")
            raise exception.BadConfigurationException(reason=message)
        self.nas_server = get_config_value("dell_nas_server")
        self.ad_domain = get_config_value("dell_ad_domain")
        self.verify_certificate = (get_config_value("dell_ssl_cert_verify") or
                                   False)
        if self.verify_certificate:
            self.certificate_path = get_config_value(
                "dell_ssl_cert_path")

        LOG.debug('Initializing Dell PowerStore REST Client.')
        LOG.info("REST server IP: %(ip)s, username: %(user)s. "
                 "Verify server's certificate: %(verify_cert)s.",
                 {
                     "ip": self.rest_ip,
                     "user": self.rest_username,
                     "verify_cert": self.verify_certificate,
                 })

        self.client = client.PowerStoreClient(self.rest_ip,
                                              self.rest_username,
                                              self.rest_password,
                                              self.verify_certificate,
                                              self.certificate_path)

        # configuration for share status update
        self.reserved_percentage = config.safe_get(
            'reserved_share_percentage')
        if self.reserved_percentage is None:
            self.reserved_percentage = 0

        self.reserved_snapshot_percentage = config.safe_get(
            'reserved_share_from_snapshot_percentage')
        if self.reserved_snapshot_percentage is None:
            self.reserved_snapshot_percentage = self.reserved_percentage

        self.reserved_share_extend_percentage = config.safe_get(
            'reserved_share_extend_percentage')
        if self.reserved_share_extend_percentage is None:
            self.reserved_share_extend_percentage = self.reserved_percentage

        self.max_over_subscription_ratio = config.safe_get(
            'max_over_subscription_ratio')

    def create_share(self, context, share, share_server):
        """Is called to create a share."""
        LOG.debug(f'Creating {share["share_proto"]} share.')
        locations = self._create_share(share)
        return locations

    def _create_share(self, share):
        """Creates a NFS or SMB share.

        In PowerStore, an export (share) belongs to a filesystem.
        This function creates a filesystem and an export.
        """
        share_name = share['name']
        size_in_bytes = share['size'] * units.Gi
        # create a filesystem
        nas_server_id = self.client.get_nas_server_id(self.nas_server)
        LOG.debug(f"Creating filesystem {share_name}")
        filesystem_id = self.client.create_filesystem(nas_server_id,
                                                      share_name,
                                                      size_in_bytes)
        if not filesystem_id:
            message = {
                _('The filesystem "%(export)s" was not created.') %
                {'export': share_name}}
            LOG.error(message)
            raise exception.ShareBackendException(msg=message)
        # create a share
        locations = self._create_share_NFS_CIFS(nas_server_id, filesystem_id,
                                                share_name,
                                                share['share_proto'].upper())
        return locations

    def _create_share_NFS_CIFS(self, nas_server_id, filesystem_id, share_name,
                               protocol):
        LOG.debug(f"Get file interfaces of {nas_server_id}")
        file_interfaces = self.client.get_nas_server_interfaces(
            nas_server_id)
        LOG.debug(f"Creating {protocol} export {share_name}")
        if protocol == 'NFS':
            export_id = self.client.create_nfs_export(filesystem_id,
                                                      share_name)
            if not export_id:
                message = (
                    _('The requested NFS export "%(export)s"'
                        ' was not created.') %
                    {'export': share_name})
                LOG.error(message)
                raise exception.ShareBackendException(msg=message)
            locations = self._get_nfs_location(file_interfaces, share_name)
        elif protocol == 'CIFS':
            export_id = self.client.create_smb_share(filesystem_id,
                                                     share_name)
            if not export_id:
                message = (
                    _('The requested SMB share "%(export)s"'
                        ' was not created.') %
                    {'export': share_name})
                LOG.error(message)
                raise exception.ShareBackendException(msg=message)
            locations = self._get_cifs_location(file_interfaces,
                                                share_name)
        return locations

    def _get_nfs_location(self, file_interfaces, share_name):
        export_locations = []
        for interface in file_interfaces:
            export_locations.append(
                {'path': f"{interface['ip']}:/{share_name}",
                 'metadata': {
                     'preferred': interface['preferred']
                     }
                 })
        return export_locations

    def _get_cifs_location(self, file_interfaces, share_name):
        export_locations = []
        for interface in file_interfaces:
            export_locations.append(
                {'path': f"\\\\{interface['ip']}\\{share_name}",
                 'metadata': {
                     'preferred': interface['preferred']
                     }
                 })
        return export_locations

    def delete_share(self, context, share, share_server):
        """Is called to delete a share."""
        LOG.debug(f'Deleting {share["share_proto"]} share.')
        self._delete_share(share)

    def _delete_share(self, share):
        """Deletes a filesystem and its associated export."""
        LOG.debug(f"Retrieving filesystem ID for filesystem {share['name']}")
        filesystem_id = self.client.get_filesystem_id(share['name'])
        if not filesystem_id:
            LOG.warning(
                f'Filesystem with share name {share["name"]} is not found.')
        else:
            LOG.debug(f"Deleting filesystem ID {filesystem_id}")
            share_deleted = self.client.delete_filesystem(filesystem_id)
            if not share_deleted:
                message = (
                    _('Failed to delete share "%(export)s".') %
                    {'export': share['name']})
                LOG.error(message)
                raise exception.ShareBackendException(msg=message)

    def extend_share(self, share, new_size, share_server):
        """Is called to extend a share."""
        LOG.debug(f"Extending {share['name']} to {new_size}GiB")
        self._resize_filesystem(share, new_size)

    def shrink_share(self, share, new_size, share_server):
        """Is called to shrink a share."""
        LOG.debug(f"Shrinking {share['name']} to {new_size}GiB")
        self._resize_filesystem(share, new_size)

    def _resize_filesystem(self, share, new_size):
        """Is called to resize a filesystem"""

        # Converts the size from GiB to Bytes
        new_size_in_bytes = new_size * units.Gi
        filesystem_id = self.client.get_filesystem_id(share['name'])
        is_success, detail = self.client.resize_filesystem(filesystem_id,
                                                           new_size_in_bytes)
        if not is_success:
            message = (_('Failed to resize share "%(export)s".') %
                       {'export': share['name']})
            LOG.error(message)
            if detail:
                raise exception.ShareShrinkingPossibleDataLoss(
                    share_id=share['id'])
            raise exception.ShareBackendException(msg=message)

    def allow_access(self, context, share, access, share_server):
        """Allow access to the share."""
        raise NotImplementedError()

    def deny_access(self, context, share, access, share_server):
        """Deny access to the share."""
        raise NotImplementedError()

    def update_access(self, context, share, access_rules, add_rules,
                      delete_rules, share_server=None):
        """Is called to update share access."""
        protocol = share['share_proto'].upper()
        LOG.debug(f'Updating access to {protocol} share.')
        if protocol == 'NFS':
            return self._update_nfs_access(share, access_rules)
        elif protocol == 'CIFS':
            return self._update_cifs_access(share, access_rules)

    def _update_nfs_access(self, share, access_rules):
        """Updates access rules for NFS share type."""
        nfs_rw_ips = set()
        nfs_ro_ips = set()
        access_updates = {}

        for rule in access_rules:
            if rule['access_type'].lower() != 'ip':
                message = (_("Only IP access type currently supported for "
                             "NFS. Share provided %(share)s with rule type "
                             "%(type)s") % {'share': share['display_name'],
                                            'type': rule['access_type']})
                LOG.error(message)
                access_updates.update({rule['access_id']: {'state': 'error'}})

            else:
                if rule['access_level'] == const.ACCESS_LEVEL_RW:
                    nfs_rw_ips.add(rule['access_to'])
                elif rule['access_level'] == const.ACCESS_LEVEL_RO:
                    nfs_ro_ips.add(rule['access_to'])
                access_updates.update({rule['access_id']: {'state': 'active'}})

        share_id = self.client.get_nfs_export_id(share['name'])
        share_updated = self.client.set_export_access(share_id,
                                                      nfs_rw_ips,
                                                      nfs_ro_ips)
        if not share_updated:
            message = (
                _('Failed to update NFS access rules for "%(export)s".') %
                {'export': share['display_name']})
            LOG.error(message)
            raise exception.ShareBackendException(msg=message)
        return access_updates

    def _update_cifs_access(self, share, access_rules):
        """Updates access rules for CIFS share type."""
        cifs_rw_users = set()
        cifs_ro_users = set()
        access_updates = {}

        for rule in access_rules:
            if rule['access_type'].lower() != 'user':
                message = (_("Only user access type currently supported for "
                             "CIFS. Share provided %(share)s with rule type "
                             "%(type)s") % {'share': share['display_name'],
                                            'type': rule['access_type']})
                LOG.error(message)
                access_updates.update({rule['access_id']: {'state': 'error'}})

            else:
                prefix = (
                    self.ad_domain or
                    self.client.get_nas_server_smb_netbios(self.nas_server)
                )
                if not prefix:
                    message = (
                        _('Failed to get daomain/netbios name of '
                          '"%(nas_server)s".'
                          ) % {'nas_server': self.nas_server})
                    LOG.error(message)
                    access_updates.update({rule['access_id']:
                                           {'state': 'error'}})
                    continue

                prefix = prefix + '\\'
                if rule['access_level'] == const.ACCESS_LEVEL_RW:
                    cifs_rw_users.add(prefix + rule['access_to'])
                elif rule['access_level'] == const.ACCESS_LEVEL_RO:
                    cifs_ro_users.add(prefix + rule['access_to'])
                access_updates.update({rule['access_id']: {'state': 'active'}})

        share_id = self.client.get_smb_share_id(share['name'])
        share_updated = self.client.set_acl(share_id,
                                            cifs_rw_users,
                                            cifs_ro_users)
        if not share_updated:
            message = (
                _('Failed to update NFS access rules for "%(export)s".') %
                {'export': share['display_name']})
            LOG.error(message)
            raise exception.ShareBackendException(msg=message)
        return access_updates

    def update_share_stats(self, stats_dict):
        """Retrieve stats info from share."""
        stats_dict['driver_version'] = VERSION
        stats_dict['storage_protocol'] = 'NFS_CIFS'
        stats_dict['reserved_percentage'] = self.reserved_percentage
        stats_dict['reserved_snapshot_percentage'] = (
            self.reserved_snapshot_percentage)
        stats_dict['reserved_share_extend_percentage'] = (
            self.reserved_share_extend_percentage)
        stats_dict['max_over_subscription_ratio'] = (
            self.max_over_subscription_ratio)

        cluster_id = self.client.get_cluster_id()
        total, used = self.client.retreive_cluster_capacity_metrics(cluster_id)
        if total and used:
            free = total - used
            stats_dict['total_capacity_gb'] = total // units.Gi
            stats_dict['free_capacity_gb'] = free // units.Gi

    def create_snapshot(self, context, snapshot, share_server):
        """Is called to create snapshot."""
        export_name = snapshot['share_name']
        LOG.debug(f'Retrieving filesystem ID for share {export_name}')
        filesystem_id = self.client.get_filesystem_id(export_name)
        if not filesystem_id:
            message = (
                _('Failed to get filesystem id for export "%(export)s".') %
                {'export': export_name})
            LOG.error(message)
            raise exception.ShareBackendException(msg=message)
        snapshot_name = snapshot['name']
        LOG.debug(
            f'Creating snapshot {snapshot_name} for filesystem {filesystem_id}'
            )
        snapshot_id = self.client.create_snapshot(filesystem_id,
                                                  snapshot_name)
        if not snapshot_id:
            message = (
                _('Failed to create snapshot "%(snapshot)s".') %
                {'snapshot': snapshot_name})
            LOG.error(message)
            raise exception.ShareBackendException(msg=message)
        else:
            LOG.info("Snapshot %(snapshot)s successfully created.",
                     {'snapshot': snapshot_name})

    def delete_snapshot(self, context, snapshot, share_server):
        """Is called to delete snapshot."""
        snapshot_name = snapshot['name']
        LOG.debug(f'Retrieving filesystem ID for snapshot {snapshot_name}')
        filesystem_id = self.client.get_filesystem_id(snapshot_name)
        LOG.debug(f'Deleting filesystem ID {filesystem_id}')
        snapshot_deleted = self.client.delete_filesystem(filesystem_id)
        if not snapshot_deleted:
            message = (
                _('Failed to delete snapshot "%(snapshot)s".') %
                {'snapshot': snapshot_name})
            LOG.error(message)
            raise exception.ShareBackendException(msg=message)
        else:
            LOG.info("Snapshot %(snapshot)s successfully deleted.",
                     {'snapshot': snapshot_name})

    def revert_to_snapshot(self, context, snapshot, share_access_rules,
                           snapshot_access_rules, share_server=None):
        """Reverts a share (in place) to the specified snapshot."""
        snapshot_name = snapshot['name']
        snapshot_id = self.client.get_filesystem_id(snapshot_name)
        snapshot_restored = self.client.restore_snapshot(snapshot_id)
        if not snapshot_restored:
            message = (
                _('Failed to restore snapshot "%(snapshot)s".') %
                {'snapshot': snapshot_name})
            LOG.error(message)
            raise exception.ShareBackendException(msg=message)
        else:
            LOG.info("Snapshot %(snapshot)s successfully restored.",
                     {'snapshot': snapshot_name})

    def create_share_from_snapshot(self, context, share, snapshot,
                                   share_server=None, parent_share=None):
        """Create a share from a snapshot - clone a snapshot."""
        LOG.debug(f'Creating {share["share_proto"]} share.')
        locations = self._create_share_from_snapshot(share, snapshot)

        if share['size'] != snapshot['size']:
            LOG.debug(f"Resizing {share['name']} to {share['size']}GiB")
            self._resize_filesystem(share, share['size'])

        return locations

    def _create_share_from_snapshot(self, share, snapshot):
        LOG.debug(f"Retrieving snapshot id of snapshot {snapshot['name']}")
        snapshot_id = self.client.get_filesystem_id(snapshot['name'])
        share_name = share['name']
        LOG.debug(
            f"Cloning filesystem {share_name} from snapshot {snapshot_id}"
            )
        filesystem_id = self.client.clone_snapshot(snapshot_id,
                                                   share_name)
        if not filesystem_id:
            message = {
                _('The filesystem "%(export)s" was not created.') %
                {'export': share_name}}
            LOG.error(message)
            raise exception.ShareBackendException(msg=message)
        # create a share
        nas_server_id = self.client.get_nas_server_id(self.nas_server)
        locations = self._create_share_NFS_CIFS(nas_server_id, filesystem_id,
                                                share_name,
                                                share['share_proto'].upper())
        return locations

    def ensure_share(self, context, share, share_server):
        """Invoked to ensure that share is exported."""

    def setup_server(self, network_info, metadata=None):
        """Set up and configures share server with given network parameters."""

    def teardown_server(self, server_details, security_services=None):
        """Teardown share server."""

    def check_for_setup_error(self):
        """Is called to check for setup error."""

    def get_default_filter_function(self):
        return 'share.size >= 3'



# Copyright (c) 2023 Dell Inc. or its subsidiaries.
# All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

"""REST client for Dell EMC PowerStore Manila Driver."""

import functools
import json

from oslo_log import log as logging
from oslo_utils import strutils
import requests

LOG = logging.getLogger(__name__)


class PowerStoreClient(object):
    def __init__(self,
                 rest_ip,
                 rest_username,
                 rest_password,
                 verify_certificate=False,
                 certificate_path=None):
        self.rest_ip = rest_ip
        self.rest_username = rest_username
        self.rest_password = rest_password
        self.verify_certificate = verify_certificate
        self.certificate_path = certificate_path
        self.base_url = "https://%s/api/rest" % self.rest_ip
        self.ok_codes = [
            requests.codes.ok,
            requests.codes.created,
            requests.codes.accepted,
            requests.codes.no_content,
            requests.codes.partial_content
        ]

    @property
    def _verify_cert(self):
        verify_cert = self.verify_certificate
        if self.verify_certificate and self.certificate_path:
            verify_cert = self.certificate_path
        return verify_cert

    def _send_request(self,
                      method,
                      url,
                      payload=None,
                      params=None,
                      log_response_data=True):
        if not params:
            params = {}
        request_params = {
            "auth": (self.rest_username, self.rest_password),
            "verify": self._verify_cert,
            "params": params
        }
        if payload and method != "GET":
            request_params["data"] = json.dumps(payload)
        request_url = self.base_url + url
        r = requests.request(method, request_url, **request_params)

        log_level = logging.DEBUG
        if r.status_code not in self.ok_codes:
            log_level = logging.ERROR
        LOG.log(log_level,
                "REST Request: %s %s with body %s",
                r.request.method,
                r.request.url,
                strutils.mask_password(r.request.body))
        if log_response_data or log_level == logging.ERROR:
            msg = "REST Response: %s with data %s" % (r.status_code, r.text)
        else:
            msg = "REST Response: %s" % r.status_code
        LOG.log(log_level, msg)

        try:
            response = r.json()
        except ValueError:
            response = None
        return r, response

    _send_get_request = functools.partialmethod(_send_request, "GET")
    _send_post_request = functools.partialmethod(_send_request, "POST")
    _send_patch_request = functools.partialmethod(_send_request, "PATCH")
    _send_delete_request = functools.partialmethod(_send_request, "DELETE")

    def get_nas_server_id(self, nas_server_name):
        """Retrieves the NAS server ID.

        :param nas_server_name: NAS server name
        :return: ID of the NAS server if success
        """
        url = '/nas_server?name=eq.' + nas_server_name
        res, response = self._send_get_request(url)
        if res.status_code == requests.codes.ok:
            return response[0]['id']

    def get_nas_server_interfaces(self, nas_server_id):
        """Retrieves the NAS server ID.

        :param nas_server_id: NAS server ID
        :return: File interfaces of the NAS server if success
        """
        url = '/nas_server/' + nas_server_id + \
            '?select=current_preferred_IPv4_interface_id,' \
            'current_preferred_IPv6_interface_id,' \
            'file_interfaces(id,ip_address)'
        res, response = self._send_get_request(url)
        if res.status_code == requests.codes.ok:
            preferred_IP = [response['current_preferred_IPv4_interface_id'],
                            response['current_preferred_IPv6_interface_id']]
            file_interfaces = []
            for i in response['file_interfaces']:
                file_interfaces.append({
                    'ip': i['ip_address'],
                    'preferred': i['id'] in preferred_IP
                })
            return file_interfaces

    def create_filesystem(self, nas_server_id, name, size):
        """Creates a filesystem.

        :param nas_server_id: ID of the nas_server
        :param name: name of the filesystem
        :param size: size in Byte
        :return: ID of the filesystem if created successfully
        """
        payload = {
            "name": name,
            "size_total": size,
            "nas_server_id": nas_server_id
            }
        url = '/file_system'
        res, response = self._send_post_request(url, payload)
        if res.status_code == requests.codes.created:
            return response["id"]

    def create_nfs_export(self, filesystem_id, name):
        """Creates an NFS export.

        :param filesystem_id: ID of the filesystem on which
                              the export will be created
        :param name: name of the NFS export
        :return: ID of the export if created successfully
        """
        payload = {
            "file_system_id": filesystem_id,
            "path": "/" + str(name),
            "name": name
            }
        url = '/nfs_export'
        res, response = self._send_post_request(url, payload)
        if res.status_code == requests.codes.created:
            return response["id"]

    def delete_filesystem(self, filesystem_id):
        """Deletes a filesystem and all associated export.

        :param filesystem_id: ID of the filesystem to delete
        :return: True if deleted successfully
        """
        url = '/file_system/' + filesystem_id
        res, _ = self._send_delete_request(url)
        return res.status_code == requests.codes.no_content

    def get_nfs_export_name(self, export_id):
        """Retrieves NFS Export name.

        :param export_id: ID of the NFS export
        :return: path of the NFS export if success
        """
        url = '/nfs_export/' + export_id + '?select=name'
        res, response = self._send_get_request(url)
        if res.status_code == requests.codes.ok:
            return response["name"]

    def get_nfs_export_id(self, name):
        """Retrieves NFS Export ID.

        :param name: name of the NFS export
        :return: id of the NFS export if success
        """
        url = '/nfs_export?select=id&name=eq.' + name
        res, response = self._send_get_request(url)
        if res.status_code == requests.codes.ok:
            return response[0]['id']

    def get_filesystem_id(self, name):
        """Retrieves an ID for a filesystem.

        :param name: name of the filesystem
        :return: ID of the filesystem if success
        """
        url = '/file_system?name=eq.' + name
        res, response = self._send_get_request(url)
        if res.status_code == requests.codes.ok:
            return response[0]['id']

    def set_export_access(self, export_id, rw_hosts, ro_hosts):
        """Sets the access hosts on the export.

        :param export_id: NFS export ID
        :param rw_hosts: a set of RW hosts
        :param ro_hosts: a set of RO hosts
        :return: True if operation succeeded
        """
        payload = {
            "read_only_hosts": list(ro_hosts),
            "read_write_root_hosts": list(rw_hosts)
            }
        url = '/nfs_export/' + export_id
        res, _ = self._send_patch_request(url, payload)
        return res.status_code == requests.codes.no_content

    def resize_filesystem(self, filesystem_id, new_size):
        """Extends the size of a share to a new size.

        :param export_id: ID of the NFS export
        :param new_size: new size to allocate in bytes
        :return: True if extended successfully
        """
        payload = {
            "size_total": new_size
            }
        url = '/file_system/' + filesystem_id
        res, response = self._send_patch_request(url, payload)
        if res.status_code == requests.codes.unprocessable and \
                response['messages'][0]['code'] == '0xE08010080449':
            return False, response['messages'][0]['message_l10n']
        return res.status_code == requests.codes.no_content, None

    def get_fsid_from_export_name(self, name):
        """Retieves the Filesystem ID used by an export.

        :param name: name of the export
        :return: ID of the Filesystem which owns the export
        """
        url = '/nfs_export?select=file_system_id&name=eq.' + name
        res, response = self._send_get_request(url)
        if res.status_code == requests.codes.ok:
            return response[0]['file_system_id']

    def create_snapshot(self, filesystem_id, name):
        """Creates a snapshot of a filesystem.

        :param filesystem_id: ID of the filesystem
        :param name: name of the snapshot
        :return: ID of the snapshot if created successfully
        """
        payload = {
            "name": name
            }
        url = '/file_system/' + filesystem_id + '/snapshot'
        res, response = self._send_post_request(url, payload)
        if res.status_code == requests.codes.created:
            return response["id"]

    def restore_snapshot(self, snapshot_id):
        """Restore a snapshot of a filesystem.

        :param snapshot_id: ID of the snapshot
        :return: True if operation succeeded
        """
        url = '/file_system/' + snapshot_id + '/restore'
        res, _ = self._send_post_request(url)
        return res.status_code == requests.codes.no_content

    def clone_snapshot(self, snapshot_id, name):
        """Clone a snapshot of a filesystem.

        :param snapshot_id: ID of the snapshot
        :param name: name the snapshot
        :return: ID of the clone if created successfully
        """
        payload = {
            "name": name
            }
        url = '/file_system/' + snapshot_id + '/clone'
        res, response = self._send_post_request(url, payload)
        if res.status_code == requests.codes.created:
            return response["id"]

    def get_cluster_id(self):
        """Get cluster id.

        :return: ID of the cluster
        """
        url = '/cluster'
        res, response = self._send_get_request(url)
        if res.status_code == requests.codes.ok:
            return response[0]["id"]

    def retreive_cluster_capacity_metrics(self, cluster_id):
        """Retreive cluster capacity metrics.

        :param cluster_id: ID of the cluster
        :return: total and used capacity in Byte
        """
        payload = {
            "entity": "space_metrics_by_cluster",
            "entity_id": cluster_id
            }
        url = '/metrics/generate?order=timestamp'
        # disable logging of the response
        res, response = self._send_post_request(url, payload,
                                                log_response_data=False)
        if res.status_code == requests.codes.ok:
            # latest cluster capacity metrics
            latestMetrics = response[len(response) - 1]
            LOG.debug(f"Latest cluster capacity: {latestMetrics}")
            return (latestMetrics["physical_total"],
                    latestMetrics["physical_used"])
        return None, None

    def create_smb_share(self, filesystem_id, name):
        """Creates a SMB share.

        :param filesystem_id: ID of the filesystem on which
                              the export will be created
        :param name: name of the SMB share
        :return: ID of the share if created successfully
        """
        payload = {
            "file_system_id": filesystem_id,
            "path": "/" + str(name),
            "name": name
            }
        url = '/smb_share'
        res, response = self._send_post_request(url, payload)
        if res.status_code == requests.codes.created:
            return response["id"]

    def get_fsid_from_share_name(self, name):
        """Retieves the Filesystem ID used by a SMB share.

        :param name: name of the SMB share
        :return: ID of the Filesystem which owns the share
        """
        url = '/smb_share?select=file_system_id&name=eq.' + name
        res, response = self._send_get_request(url)
        if res.status_code == requests.codes.ok:
            return response[0]['file_system_id']

    def get_smb_share_id(self, name):
        """Retrieves SMB share ID.

        :param name: name of the SMB share
        :return: id of the SMB share if success
        """
        url = '/smb_share?select=id&name=eq.' + name
        res, response = self._send_get_request(url)
        if res.status_code == requests.codes.ok:
            return response[0]['id']

    def get_nas_server_smb_netbios(self, nas_server_name):
        """Retrieves the domain name or netbios name.

        :param nas_server_name: NAS server name
        :return: Netbios name of SMB server if success
        """
        url = '/nas_server?select=smb_servers(is_standalone,netbios_name)' \
            '&name=eq.' + nas_server_name
        res, response = self._send_get_request(url)
        if res.status_code == requests.codes.ok:
            smb_server = response[0]['smb_servers'][0]
            if smb_server["is_standalone"]:
                return smb_server["netbios_name"]

    def set_acl(self, smb_share_id, cifs_rw_users, cifs_ro_users):
        """Set ACL for a SMB share.

        :param smb_share_id: ID of the SMB share
        :param name: name of the SMB share
        :return: ID of the share if created successfully
        """
        aces = list()
        for rw_user in cifs_rw_users:
            ace = {
                "trustee_type": "User",
                "trustee_name": rw_user,
                "access_level": "Change",
                "access_type": "Allow"
            }
            aces.append(ace)

        for ro_user in cifs_ro_users:
            ace = {
                "trustee_type": "User",
                "trustee_name": ro_user,
                "access_level": "Read",
                "access_type": "Allow"
            }
            aces.append(ace)

        payload = {
            "aces": aces
            }
        url = '/smb_share/' + smb_share_id + '/set_acl'
        res, _ = self._send_post_request(url, payload)
        return res.status_code == requests.codes.no_content



# Copyright 2021 Red Hat, Inc
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""
Helpers for lvm related routines
"""

from oslo_concurrency import processutils

import manila.privsep


@manila.privsep.sys_admin_pctxt.entrypoint
def lvremove(vg_name, lv_name):
    processutils.execute('lvremove', '-f', f'{vg_name}/{lv_name}')


@manila.privsep.sys_admin_pctxt.entrypoint
def lvcreate(lv_size, lv_name, vg_name, mirrors=0, region_size=0):
    extra_params = []

    if mirrors:
        extra_params += ['-m', mirrors, '--nosync']
    if region_size:
        extra_params += ['-R', region_size]

    processutils.execute(
        'lvcreate', '-Wy', '--yes', '-L', f'{lv_size}G', '-n', lv_name,
        vg_name, *extra_params)


@manila.privsep.sys_admin_pctxt.entrypoint
def lv_snapshot_create(snapshot_size, snap_name, orig_lv_name):
    size_str = '%sG' % snapshot_size
    processutils.execute(
        'lvcreate', '-L', size_str, '--name', snap_name,
        '--snapshot', orig_lv_name)


@manila.privsep.sys_admin_pctxt.entrypoint
def get_vgs(vg_name):
    out, err = processutils.execute(
        'vgs', vg_name, '--rows', '--units', 'g',)
    return out, err


@manila.privsep.sys_admin_pctxt.entrypoint
def list_vgs_get_name():
    out, err = processutils.execute('vgs', '--noheadings', '-o', 'name')
    return out, err


@manila.privsep.sys_admin_pctxt.entrypoint
def lvconvert(vg_name, snapshot_name):
    processutils.execute(
        'lvconvert', '--merge', f'{vg_name}/{snapshot_name}')


@manila.privsep.sys_admin_pctxt.entrypoint
def lvrename(vg_name, lv_name, new_name):
    processutils.execute(
        'lvrename', vg_name, lv_name, new_name)


@manila.privsep.sys_admin_pctxt.entrypoint
def lvextend(lv_name, new_size):
    processutils.execute('lvextend', '-L', '%sG' % new_size, '-r', lv_name)



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

"""add_resource_locks

Revision ID: cb20f743ca7b
Revises: 9afbe2df4945
Create Date: 2023-06-23 16:34:36.277477

"""

# revision identifiers, used by Alembic.
revision = 'cb20f743ca7b'
down_revision = '9afbe2df4945'

from alembic import op
from oslo_log import log
import sqlalchemy as sa

LOG = log.getLogger(__name__)


def upgrade():
    context = op.get_context()
    mysql_dl = context.bind.dialect.name == 'mysql'
    datetime_type = (sa.dialects.mysql.DATETIME(fsp=6)
                     if mysql_dl else sa.DateTime)
    try:
        op.create_table(
            'resource_locks',
            sa.Column('id', sa.String(36), primary_key=True, nullable=False),
            sa.Column('user_id', sa.String(255), nullable=False),
            sa.Column('project_id', sa.String(255), nullable=False),
            sa.Column('resource_action', sa.String(255), default='delete'),
            sa.Column('resource_type', sa.String(255), nullable=False),
            sa.Column('resource_id', sa.String(36), nullable=False),
            sa.Column('lock_context', sa.String(16), nullable=False),
            sa.Column('lock_reason', sa.String(1023), nullable=True),
            sa.Column('created_at', datetime_type),
            sa.Column('updated_at', datetime_type),
            sa.Column('deleted_at', datetime_type),
            sa.Column('deleted', sa.String(36), default='False'),
            mysql_engine='InnoDB',
            mysql_charset='utf8',
        )
    except Exception:
        LOG.error("Table resource_locks not created!")
        raise


def downgrade():
    try:
        op.drop_table('resource_locks')
    except Exception:
        LOG.error("resource_locks table not dropped")
        raise



# Copyright (c) 2015 Mirantis inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from manila.api import common


class ViewBuilder(common.ViewBuilder):

    _collection_name = "services"
    _detail_version_modifiers = [
        "add_disabled_reason_field",
    ]

    def summary(self, request, service):
        """Summary view of a single service."""
        keys = 'host', 'binary', 'status',
        service_dict = {key: service.get(key) for key in keys}
        self.update_versioned_resource_dict(request, service_dict, service)
        return service_dict

    def detail(self, request, service):
        """Detailed view of a single service."""
        keys = ('id', 'binary', 'host', 'zone', 'status',
                'state', 'updated_at')
        service_dict = {key: service.get(key) for key in keys}
        self.update_versioned_resource_dict(request, service_dict, service)
        return service_dict

    def detail_list(self, request, services):
        """Detailed view of a list of services."""
        services_list = [self.detail(request, s) for s in services]
        services_dict = dict(services=services_list)
        return services_dict

    @common.ViewBuilder.versioned_method("2.83")
    def add_disabled_reason_field(self, context, service_dict, service):
        service_dict.pop('disabled', None)
        service_dict['status'] = service.get('status')
        service_dict['disabled_reason'] = service.get('disabled_reason')



# Copyright 2012 NetApp
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
"""Unit tests for the LVM driver module."""

import os
from unittest import mock

import ddt
from oslo_concurrency import processutils
from oslo_config import cfg
from oslo_utils import timeutils

from manila.common import constants as const
from manila import context
from manila import exception
from manila.privsep import common as privsep_common
from manila.privsep import filesystem
from manila.privsep import lvm as privsep_lvm
from manila.privsep import os as os_routines
from manila.share import configuration
from manila.share.drivers import lvm
from manila import test
from manila.tests.db import fakes as db_fakes
from manila.tests import fake_utils
from manila.tests.share.drivers import test_generic


CONF = cfg.CONF


def fake_share(**kwargs):
    share = {
        'id': 'fakeid',
        'name': 'fakename',
        'size': 1,
        'share_proto': 'NFS',
        'export_location': '127.0.0.1:/mnt/nfs/volume-00002',
    }
    share.update(kwargs)
    return db_fakes.FakeModel(share)


def fake_snapshot(**kwargs):
    snapshot = {
        'id': 'fakesnapshotid',
        'share_name': 'fakename',
        'share_id': 'fakeid',
        'name': 'fakesnapshotname',
        'share_proto': 'NFS',
        'export_location': '127.0.0.1:/mnt/nfs/volume-00002',
        'share': {
            'id': 'fakeid',
            'name': 'fakename',
            'size': 1,
            'share_proto': 'NFS',
        },
    }
    snapshot.update(kwargs)
    return db_fakes.FakeModel(snapshot)


def fake_access(**kwargs):
    access = {
        'id': 'fakeaccid',
        'access_type': 'ip',
        'access_to': '10.0.0.2',
        'access_level': 'rw',
        'state': 'active',
    }
    access.update(kwargs)
    return db_fakes.FakeModel(access)


@ddt.ddt
class LVMShareDriverTestCase(test.TestCase):
    """Tests LVMShareDriver."""

    def setUp(self):
        super(LVMShareDriverTestCase, self).setUp()
        fake_utils.stub_out_utils_execute(self)
        self._context = context.get_admin_context()

        CONF.set_default('lvm_share_volume_group', 'fakevg')
        CONF.set_default('lvm_share_export_ips', ['10.0.0.1', '10.0.0.2'])
        CONF.set_default('driver_handles_share_servers', False)
        CONF.set_default('reserved_share_percentage', 50)
        CONF.set_default('reserved_share_from_snapshot_percentage', 30)
        CONF.set_default('reserved_share_extend_percentage', 30)

        self._helper_cifs = mock.Mock()
        self._helper_nfs = mock.Mock()
        self.fake_conf = configuration.Configuration(None)
        self._db = mock.Mock()
        self._os = lvm.os = mock.Mock()
        self._os.path.join = os.path.join
        self._driver = lvm.LVMShareDriver(self._db,
                                          configuration=self.fake_conf)
        self._driver._helpers = {
            'CIFS': self._helper_cifs,
            'NFS': self._helper_nfs,
        }

        self.share = fake_share()
        self.access = fake_access()
        self.snapshot = fake_snapshot()
        self.server = {
            'public_addresses': self.fake_conf.lvm_share_export_ips,
            'instance_id': 'LVM',
            'lock_name': 'manila_lvm',
        }

        # Used only to test compatibility with share manager
        self.share_server = "fake_share_server"

    def tearDown(self):
        super(LVMShareDriverTestCase, self).tearDown()
        fake_utils.fake_execute_set_repliers([])
        fake_utils.fake_execute_clear_log()

    def test_do_setup(self):
        CONF.set_default('lvm_share_helpers', ['NFS=fakenfs'])
        lvm.importutils = mock.Mock()
        lvm.importutils.import_class.return_value = self._helper_nfs
        self._driver.do_setup(self._context)
        lvm.importutils.import_class.assert_has_calls([
            mock.call('fakenfs')
        ])

    def test_check_for_setup_error(self):
        out, err = '\n   fake1\n   fakevg\n   fake2\n', ''
        self.mock_object(privsep_lvm, 'list_vgs_get_name',
                         mock.Mock(return_value=(out, err)))

        self._driver.check_for_setup_error()

        privsep_lvm.list_vgs_get_name.assert_called_once()

    def test_check_for_setup_error_no_vg(self):
        out = '\n   fake0\n   fake1\n   fake2\n'
        err = ''
        self.mock_object(privsep_lvm, 'list_vgs_get_name',
                         mock.Mock(return_value=(out, err)))
        self.assertRaises(exception.InvalidParameterValue,
                          self._driver.check_for_setup_error)

    def test_check_for_setup_error_no_export_ips(self):
        out = '\n   fake1\n   fakevg\n   fake2\n'
        err = ''
        self.mock_object(privsep_lvm, 'list_vgs_get_name',
                         mock.Mock(return_value=(out, err)))

        CONF.set_default('lvm_share_export_ips', None)
        self.assertRaises(exception.InvalidParameterValue,
                          self._driver.check_for_setup_error)

    def test_local_path_normal(self):
        share = fake_share(name='fake_sharename')
        CONF.set_default('lvm_share_volume_group', 'fake_vg')
        ret = self._driver._get_local_path(share)
        self.assertEqual('/dev/mapper/fake_vg-fake_sharename', ret)

    def test_local_path_escapes(self):
        share = fake_share(name='fake-sharename')
        CONF.set_default('lvm_share_volume_group', 'fake-vg')
        ret = self._driver._get_local_path(share)
        self.assertEqual('/dev/mapper/fake--vg-fake--sharename', ret)

    def test_create_share(self):
        CONF.set_default('lvm_share_mirrors', 0)
        self._driver._mount_device = mock.Mock()
        lv_create_mock = privsep_lvm.lvcreate = mock.Mock()
        lv_create_args = [
            self.share['size'], self.share['name'],
            CONF.lvm_share_volume_group, 0, 0]
        self.mock_object(privsep_common, 'execute_with_retries')
        self.mock_object(filesystem, 'make_filesystem')
        self.mock_object(self._driver, '_get_mount_point_name',
                         mock.Mock(return_value=self.share['name']))

        ret = self._driver.create_share(self._context, self.share,
                                        self.share_server)

        self._driver._mount_device.assert_called_with(
            self.share, '/dev/mapper/fakevg-fakename')
        privsep_common.execute_with_retries.assert_called_once_with(
            lv_create_mock, lv_create_args, CONF.num_shell_tries)
        filesystem.make_filesystem.assert_called_once_with(
            'ext4', '/dev/mapper/fakevg-fakename')
        self.assertEqual(self._helper_nfs.create_exports.return_value, ret)

    def test_create_share_from_snapshot(self):
        CONF.set_default('lvm_share_mirrors', 0)
        self._driver._mount_device = mock.Mock()
        snapshot_instance = {
            'snapshot_id': 'fakesnapshotid',
            'name': 'fakename'
        }
        mount_share = '/dev/mapper/fakevg-fakename'
        mount_snapshot = '/dev/mapper/fakevg-fakename'
        self._helper_nfs.create_export.return_value = 'fakelocation'
        lv_create_mock = privsep_lvm.lvcreate = mock.Mock()
        lv_create_args = [
            self.share['size'], self.share['name'],
            CONF.lvm_share_volume_group, 0, 0]

        self.mock_object(privsep_common, 'execute_with_retries')
        self.mock_object(os_routines, 'is_data_definition_direct_io_supported',
                         mock.Mock(return_value=True))
        self.mock_object(os_routines, 'data_definition')
        self.mock_object(os_routines, 'mount')
        self.mock_object(os_routines, 'chmod')
        self.mock_object(filesystem, 'make_filesystem')
        self.mock_object(filesystem, 'e2fsck')
        self.mock_object(filesystem, 'tune2fs')
        self.mock_object(self._driver, '_get_mount_point_name',
                         mock.Mock(return_value=self.share['name']))

        self._driver.create_share_from_snapshot(self._context,
                                                self.share,
                                                snapshot_instance,
                                                self.share_server)

        self._driver._mount_device.assert_called_with(self.share,
                                                      mount_snapshot)
        privsep_common.execute_with_retries.assert_called_once_with(
            lv_create_mock, lv_create_args, 3)
        filesystem.make_filesystem.assert_called_once_with(
            'ext4', '/dev/mapper/fakevg-fakename')
        filesystem.e2fsck.assert_called_once_with(
            mount_share)
        filesystem.tune2fs.assert_called_once_with(
            mount_share)
        (os_routines.is_data_definition_direct_io_supported
            .assert_called_once_with(
                mount_snapshot, mount_share))
        os_routines.data_definition.assert_called_once_with(
            mount_snapshot, mount_share, (self.share['size'] * 1024),
            use_direct_io=True)

    def test_create_share_mirrors(self):
        share = fake_share(size='2048')
        CONF.set_default('lvm_share_mirrors', 2)
        lv_create_mock = privsep_lvm.lvcreate = mock.Mock()
        lv_create_args = [
            '2048', self.share['name'],
            CONF.lvm_share_volume_group, 2, '2']

        self._driver._mount_device = mock.Mock()
        self.mock_object(privsep_common, 'execute_with_retries')
        self.mock_object(filesystem, 'make_filesystem')
        self.mock_object(self._driver, '_get_mount_point_name',
                         mock.Mock(return_value=self.share['name']))

        ret = self._driver.create_share(self._context, share,
                                        self.share_server)

        self._driver._mount_device.assert_called_with(
            share, '/dev/mapper/fakevg-fakename')
        privsep_common.execute_with_retries.assert_called_once_with(
            lv_create_mock, lv_create_args, 3)
        filesystem.make_filesystem.assert_called_once_with(
            'ext4', '/dev/mapper/fakevg-fakename')
        self.assertEqual(self._helper_nfs.create_exports.return_value, ret)

    def test_deallocate_container(self):
        mock_lvremove = privsep_lvm.lvremove = mock.Mock()
        self.mock_object(privsep_common, 'execute_with_retries')

        self._driver._deallocate_container(self.share['name'])

        privsep_common.execute_with_retries.assert_called_once_with(
            mock_lvremove, [CONF.lvm_share_volume_group, self.share['name']], 3
        )

    def test_deallocate_container_error(self):
        def _fake_exec(*args, **kwargs):
            raise exception.ProcessExecutionError(stderr="error")

        self.mock_object(privsep_common, 'execute_with_retries', _fake_exec)
        self.assertRaises(exception.ProcessExecutionError,
                          self._driver._deallocate_container,
                          self.share['name'])

    @ddt.data(
        'Logical volume "fake/fake-volume" not found\n',
        'Failed to find logical volume "fake/fake-volume"\n')
    def test_deallocate_container_not_found_error(self, error_msg):
        def _fake_exec(*args, **kwargs):
            raise exception.ProcessExecutionError(stderr=error_msg)

        self.mock_object(privsep_common, 'execute_with_retries', _fake_exec)
        self._driver._deallocate_container(self.share['name'])

    @mock.patch.object(lvm.LVMShareDriver, '_update_share_stats', mock.Mock())
    def test_get_share_stats(self):
        with mock.patch.object(self._driver, '_stats', mock.Mock) as stats:
            self.assertEqual(stats, self._driver.get_share_stats())
        self.assertFalse(self._driver._update_share_stats.called)

    @mock.patch.object(lvm.LVMShareDriver, '_update_share_stats', mock.Mock())
    def test_get_share_stats_refresh(self):
        with mock.patch.object(self._driver, '_stats', mock.Mock) as stats:
            self.assertEqual(stats,
                             self._driver.get_share_stats(refresh=True))
        self._driver._update_share_stats.assert_called_once_with()

    def test__unmount_device_not_mounted(self):
        mount_path = self._get_mount_path(self.share)
        error_msg = (
            "umount: /opt/stack/data/manila/mnt/share-fake-share: not "
            "mounted.\n")
        umount_exception = exception.ProcessExecutionError(stderr=error_msg)
        self.mock_object(
            os_routines, 'umount', mock.Mock(side_effect=umount_exception))
        self.mock_object(os_routines, 'rmdir')
        self._os.path.exists.return_value = True

        self._driver._unmount_device(self.share, raise_if_missing=False)

        self._os.path.exists.assert_called_with(mount_path)
        os_routines.umount.assert_called_once_with(mount_path)

    def test__unmount_device_is_busy_error(self):
        error_msg = 'device is busy'
        umount_exception = exception.ProcessExecutionError(stderr=error_msg)
        self.mock_object(
            os_routines, 'umount', mock.Mock(side_effect=umount_exception))
        self._os.path.exists.return_value = True
        mount_path = self._get_mount_path(self.share)

        self.assertRaises(exception.ShareBusyException,
                          self._driver._unmount_device,
                          self.share)
        os_routines.umount.assert_called_once_with(mount_path)

    def test__unmount_device_error(self):
        error_msg = 'fake error'
        mount_path = self._get_mount_path(self.share)
        umount_exception = exception.ProcessExecutionError(stderr=error_msg)
        self.mock_object(
            os_routines, 'umount', mock.Mock(side_effect=umount_exception))
        self._os.path.exists.return_value = True

        self.assertRaises(processutils.ProcessExecutionError,
                          self._driver._unmount_device,
                          self.share)
        self._os.path.exists.assert_called_with(mount_path)
        os_routines.umount.assert_called_once_with(mount_path)

    def test__unmount_device_rmdir_error(self):
        error_msg = 'fake error'
        mount_path = self._get_mount_path(self.share)
        umount_exception = exception.ProcessExecutionError(stderr=error_msg)
        self.mock_object(os_routines, 'umount')
        self.mock_object(os_routines, 'rmdir',
                         mock.Mock(side_effect=umount_exception))
        self._os.path.exists.return_value = True

        self.assertRaises(exception.ShareBackendException,
                          self._driver._unmount_device,
                          self.share)
        self._os.path.exists.assert_called_with(mount_path)
        os_routines.umount.assert_called_once_with(mount_path)
        os_routines.rmdir.assert_called_once_with(mount_path)

    def test_create_snapshot(self):
        mock_lv_create = privsep_lvm.lvcreate = mock.Mock()
        orig_lv_name = "%s/%s" % (CONF.lvm_share_volume_group,
                                  self.snapshot['share_name'])
        device_path = '/dev/mapper/fakevg-%s' % self.snapshot['name']
        lv_create_args = [
            self.snapshot['share']['size'], self.snapshot['share']['name'],
            orig_lv_name]

        self.mock_object(privsep_common, 'execute_with_retries')
        self.mock_object(filesystem, 'e2fsck')
        self.mock_object(filesystem, 'tune2fs')
        self.mock_object(os_routines, 'mount')
        self.mock_object(os_routines, 'chmod')

        self._driver.create_snapshot(self._context, self.snapshot,
                                     self.share_server)
        mount_path = self._get_mount_path(self.snapshot)
        expected_exec = [
            "mkdir -p " + mount_path,
        ]
        self.assertEqual(expected_exec, fake_utils.fake_execute_get_log())
        privsep_common.execute_with_retries(
            mock_lv_create, lv_create_args, CONF.num_shell_tries)
        filesystem.e2fsck.assert_called_once_with(device_path)
        filesystem.tune2fs.assert_called_once_with(device_path)
        os_routines.mount.assert_called_once_with(
            "/dev/mapper/fakevg-fakesnapshotname", mount_path)
        os_routines.chmod.assert_called_once_with(
            '777', mount_path)

    def test_ensure_share(self):
        device_name = '/dev/mapper/fakevg-fakename'
        self.mock_object(self._driver, '_get_mount_point_name',
                         mock.Mock(return_value=self.share['name']))
        with mock.patch.object(self._driver,
                               '_mount_device',
                               mock.Mock(return_value='fake_location')):
            self._driver.ensure_share(self._context, self.share,
                                      self.share_server)
            self._driver._mount_device.assert_called_with(self.share,
                                                          device_name)
            self._helper_nfs.create_exports.assert_called_once_with(
                self.server, self.share['name'], recreate=True)

    def test_delete_share(self):
        self.mock_object(self._driver, '_get_mount_point_name',
                         mock.Mock(return_value=self.share['name']))
        mount_path = self._get_mount_path(self.share)
        self._helper_nfs.remove_export(mount_path, self.share['name'])
        self._driver._delete_share(self._context, self.share)

    def test_delete_snapshot(self):
        mount_path = self._get_mount_path(self.snapshot)
        self.mock_object(os_routines, 'umount')
        self.mock_object(os_routines, 'rmdir')
        self.mock_object(privsep_common, 'execute_with_retries')
        self.mock_object(self._driver, '_deallocate_container')

        self._driver.delete_snapshot(self._context, self.snapshot,
                                     self.share_server)

        os_routines.umount.assert_called_once_with(mount_path)
        os_routines.rmdir.assert_called_once_with(mount_path)
        self._driver._deallocate_container.assert_called_once_with(
            self.snapshot['name'])

    def test_delete_share_invalid_share(self):
        self.mock_object(self._driver, '_unmount_device')
        self.mock_object(self._driver, '_deallocate_container')
        self._driver._get_helper = mock.Mock(
            side_effect=exception.InvalidShare(reason='fake'))
        self.mock_object(self._driver, '_get_mount_point_name',
                         mock.Mock(return_value=self.share['name']))

        self._driver.delete_share(self._context, self.share, self.share_server)

        self._driver._unmount_device.assert_called_once_with(
            self.share, raise_if_missing=False, retry_busy_device=True)
        self._driver._deallocate_container.assert_called_once_with(
            self.share['name'])

    def test_delete_share_process_execution_error(self):
        self.mock_object(
            self._helper_nfs,
            'remove_export',
            mock.Mock(side_effect=exception.ProcessExecutionError))
        self.mock_object(self._driver, '_get_mount_point_name',
                         mock.Mock(return_value=self.share['name']))

        self._driver._delete_share(self._context, self.share)
        self._helper_nfs.remove_exports.assert_called_once_with(
            self.server,
            self.share['name'])

    @ddt.data(const.ACCESS_LEVEL_RW, const.ACCESS_LEVEL_RO)
    def test_update_access(self, access_level):
        access_rules = [test_generic.get_fake_access_rule(
            '1.1.1.1', access_level), ]
        add_rules = [test_generic.get_fake_access_rule(
            '2.2.2.2', access_level), ]
        delete_rules = [test_generic.get_fake_access_rule(
            '3.3.3.3', access_level), ]
        self.mock_object(self._driver, '_get_mount_point_name',
                         mock.Mock(return_value=self.share['name']))
        self._driver.update_access(self._context, self.share, access_rules,
                                   add_rules=add_rules,
                                   delete_rules=delete_rules,
                                   share_server=self.server)
        (self._driver._helpers[self.share['share_proto']].
            update_access.assert_called_once_with(
                self.server, self.share['name'],
                access_rules, add_rules=add_rules, delete_rules=delete_rules))

    @ddt.data((['1001::1001/129'], False),
              (['1.1.1.256'], False),
              (['1001::1001'], [6]),
              ('1.1.1.0', [4]),
              (['1001::1001', '1.1.1.0'], [6, 4]),
              (['1001::1001/129', '1.1.1.0'], False))
    @ddt.unpack
    def test_get_configured_ip_versions(self, configured_ips,
                                        configured_ip_version):
        CONF.set_default('lvm_share_export_ips', configured_ips)
        if configured_ip_version:
            self.assertEqual(configured_ip_version,
                             self._driver.get_configured_ip_versions())
        else:
            self.assertRaises(exception.InvalidInput,
                              self._driver.get_configured_ip_versions)

    def test_mount_device(self):
        mount_path = self._get_mount_path(self.share)
        self.mock_object(os_routines, 'mount')
        self.mock_object(os_routines, 'chmod')
        expected_exec = [
            "mkdir -p %s" % (mount_path,),
        ]
        device_name = 'fakedevice'

        ret = self._driver._mount_device(self.share, device_name)

        self.assertEqual(expected_exec, fake_utils.fake_execute_get_log())
        os_routines.mount.assert_called_once_with(device_name, mount_path)
        os_routines.chmod.assert_called_once_with('777', mount_path)
        self.assertEqual(mount_path, ret)

    def test_mount_device_already(self):
        def exec_runner(*args, **kwargs):
            if 'mount' in args and '-l' not in args:
                raise exception.ProcessExecutionError()
            else:
                return 'fakedevice', ''

        self.mock_object(self._driver, '_execute', exec_runner)
        self.mock_object(os_routines, 'mount')
        self.mock_object(os_routines, 'chmod')
        mount_path = self._get_mount_path(self.share)

        ret = self._driver._mount_device(self.share, 'fakedevice')
        self.assertEqual(mount_path, ret)

    def test_mount_device_error(self):
        self.mock_object(self._driver, '_execute')
        self.mock_object(
            os_routines, 'mount',
            mock.Mock(side_effect=exception.ProcessExecutionError))
        self.mock_object(
            os_routines, 'list_mounts',
            mock.Mock(return_value=('fake', '')))

        self.assertRaises(exception.ProcessExecutionError,
                          self._driver._mount_device, self.share, 'fakedevice')

    def test_get_helper(self):
        share_cifs = fake_share(share_proto='CIFS')
        share_nfs = fake_share(share_proto='NFS')
        share_fake = fake_share(share_proto='FAKE')
        self.assertEqual(self._driver._get_helper(share_cifs),
                         self._helper_cifs)
        self.assertEqual(self._driver._get_helper(share_nfs),
                         self._helper_nfs)
        self.assertRaises(exception.InvalidShare, self._driver._get_helper,
                          share_fake)

    def _get_mount_path(self, share):
        return os.path.join(CONF.lvm_share_export_root, share['name'])

    @ddt.data(True, False)
    def test__unmount_device_with_retry_busy_device(self, retry_busy_device):
        execute_sideeffects = [
            exception.ProcessExecutionError(stderr='device is busy'),
            exception.ProcessExecutionError(stderr='target is busy'),
            None, None
        ] if retry_busy_device else [None, None]
        mount_path = self._get_mount_path(self.share)
        self._os.path.exists.return_value = True
        self.mock_object(os_routines, 'umount', mock.Mock(
            side_effect=execute_sideeffects))
        self.mock_object(os_routines, 'rmdir')

        self._driver._unmount_device(self.share,
                                     retry_busy_device=retry_busy_device)

        num_of_times_umount_is_called = 3 if retry_busy_device else 1

        self._os.path.exists.assert_called_with(mount_path)
        os_routines.umount.assert_has_calls([
            mock.call(mount_path)] * num_of_times_umount_is_called)
        os_routines.rmdir.assert_called_once_with(mount_path)

    def test_extend_share(self):
        local_path = self._driver._get_local_path(self.share)
        self.mock_object(self._driver, '_extend_container')
        self.mock_object(self._driver, '_execute')
        self._driver.extend_share(self.share, 3)
        self._driver._extend_container.assert_called_once_with(self.share,
                                                               local_path, 3)

    def test_ssh_exec_as_root(self):
        command = ['fake_command']
        self.mock_object(self._driver, '_execute')
        self._driver._ssh_exec_as_root('fake_server', command)
        self._driver._execute.assert_called_once_with('fake_command',
                                                      check_exit_code=True)

    def test_ssh_exec_as_root_with_sudo(self):
        command = ['sudo', 'fake_command']
        self.mock_object(self._driver, '_execute')
        self._driver._ssh_exec_as_root('fake_server', command)
        self._driver._execute.assert_called_once_with(
            'fake_command', run_as_root=True, check_exit_code=True)

    def test_extend_container(self):
        mock_lvextend = privsep_lvm.lvextend = mock.Mock()

        self.mock_object(privsep_common, 'execute_with_retries')
        self._driver._extend_container(self.share, 'device_name', 3)

        privsep_common.execute_with_retries.assert_called_once_with(
            mock_lvextend, ['device_name', 3], CONF.num_shell_tries)

    def test_get_share_server_pools(self):
        expected_result = [{
            'pool_name': 'lvm-single-pool',
            'total_capacity_gb': 33,
            'free_capacity_gb': 22,
            'reserved_percentage': 0,
            'reserved_snapshot_percentage': 0,
            'reserved_share_extend_percentage': 0,
            'mount_point_name_support': True,
        }, ]
        out, err = "VSize 33g VFree 22g", None
        self.mock_object(
            privsep_lvm, 'get_vgs', mock.Mock(return_value=(out, err)))

        self.assertEqual(expected_result,
                         self._driver.get_share_server_pools())

    @ddt.data(True, False)
    def test_copy_volume_error(self, use_direct_io):
        src_str = 'src'
        dest_str = 'dest'
        self.mock_object(
            os_routines, 'is_data_definition_direct_io_supported',
            mock.Mock(return_value=use_direct_io))
        self.mock_object(
            os_routines, 'data_definition',
            mock.Mock(side_effect=exception.ProcessExecutionError))

        self.assertRaises(
            exception.ShareBackendException,
            self._driver._copy_volume, src_str, dest_str, 1)
        (os_routines.is_data_definition_direct_io_supported
            .assert_called_once_with(
                src_str, dest_str))
        os_routines.data_definition.assert_called_once_with(
            src_str, dest_str, (1 * 1024), use_direct_io=use_direct_io)

    @ddt.data((['1.1.1.1'], 4), (['1001::1001'], 6))
    @ddt.unpack
    def test_update_share_stats(self, configured_ip, version):
        CONF.set_default('lvm_share_export_ips', configured_ip)
        self.mock_object(self._driver, 'get_share_server_pools',
                         mock.Mock(return_value='test-pool'))

        self._driver._update_share_stats()
        self.assertEqual('LVM', self._driver._stats['share_backend_name'])
        self.assertEqual('NFS_CIFS', self._driver._stats['storage_protocol'])
        self.assertEqual(50, self._driver._stats['reserved_percentage'])
        self.assertTrue(self._driver._stats['snapshot_support'])
        self.assertEqual('LVMShareDriver', self._driver._stats['driver_name'])
        self.assertEqual('test-pool', self._driver._stats['pools'])
        self.assertEqual(version == 4, self._driver._stats['ipv4_support'])
        self.assertEqual(version == 6, self._driver._stats['ipv6_support'])

    def test_revert_to_snapshot(self):
        share_local_path = '/dev/mapper/fakevg-fakename'
        snapshot_local_path = '/dev/mapper/fakevg-fakesnapshotname'
        mock_update_access = self.mock_object(
            self._helper_nfs, 'update_access')
        mock__unmount_device = self.mock_object(
            self._driver, '_unmount_device')
        mock_lvconvert = self.mock_object(privsep_lvm, 'lvconvert')
        mock_create_snapshot = self.mock_object(
            self._driver, '_create_snapshot')
        mock_mount_device = self.mock_object(
            self._driver, '_mount_device')
        mock_get_local_path = self.mock_object(
            self._driver, '_get_local_path',
            mock.Mock(side_effect=[share_local_path, snapshot_local_path]))
        self.mock_object(self._driver, '_get_mount_point_name',
                         mock.Mock(return_value=self.snapshot['name']))
        snapshot_parent_share = self.snapshot['share']

        self._driver.revert_to_snapshot(self._context, self.snapshot,
                                        [], [], self.share_server)
        self.assertEqual(4, mock_update_access.call_count)
        mock__unmount_device.assert_has_calls(
            [mock.call(self.snapshot), mock.call(self.snapshot['share'])])
        mock_lvconvert.assert_called_once_with(
            CONF.lvm_share_volume_group, self.snapshot['name'])
        mock_create_snapshot.assert_called_once_with(
            self._context, self.snapshot)
        mock_mount_device.assert_has_calls(
            [mock.call(snapshot_parent_share, share_local_path),
             mock.call(self.snapshot, snapshot_local_path)]
        )
        mock_get_local_path.assert_has_calls(
            [mock.call(snapshot_parent_share),
             mock.call(self.snapshot)])

    def test_snapshot_update_access(self):
        access_rules = [{
            'access_type': 'ip',
            'access_to': '1.1.1.1',
            'access_level': 'ro',
        }]

        add_rules = [{
            'access_type': 'ip',
            'access_to': '2.2.2.2',
            'access_level': 'ro',
        }]

        delete_rules = [{
            'access_type': 'ip',
            'access_to': '3.3.3.3',
            'access_level': 'ro',
        }]

        self._driver.snapshot_update_access(self._context, self.snapshot,
                                            access_rules, add_rules,
                                            delete_rules)

        (self._driver._helpers[self.snapshot['share']['share_proto']].
            update_access.assert_called_once_with(
            self.server, self.snapshot['name'],
            access_rules, add_rules=add_rules, delete_rules=delete_rules))

    @mock.patch.object(timeutils, 'utcnow', mock.Mock(
                       return_value='fake_date'))
    def test_update_share_usage_size(self):
        mount_path = self._get_mount_path(self.share)
        self._os.path.exists.return_value = True
        self.mock_object(
            self._driver,
            '_execute',
            mock.Mock(return_value=(
                "Mounted on                    Used "
                + mount_path + "               1G", None)))

        update_shares = self._driver.update_share_usage_size(
            self._context, [self.share, ])
        self._os.path.exists.assert_called_with(mount_path)
        self.assertEqual(
            [{'id': 'fakeid', 'used_size': '1',
              'gathered_at': 'fake_date'}],
            update_shares)
        self._driver._execute.assert_called_once_with(
            'df', '-l', '--output=target,used',
            '--block-size=g')

    @mock.patch.object(timeutils, 'utcnow', mock.Mock(
                       return_value='fake_date'))
    def test_update_share_usage_size_multiple_share(self):
        share1 = fake_share(id='fakeid_get_fail', name='get_fail')
        share2 = fake_share(id='fakeid_success', name='get_success')
        share3 = fake_share(id='fakeid_not_exist', name='get_not_exist')

        mount_path2 = self._get_mount_path(share2)
        mount_path3 = self._get_mount_path(share3)
        self._os.path.exists.side_effect = [True, True, False]
        self.mock_object(
            self._driver,
            '_execute',
            mock.Mock(return_value=(
                "Mounted on                    Used "
                + mount_path2 + "               1G", None)))

        update_shares = self._driver.update_share_usage_size(
            self._context, [share1, share2, share3])
        self._os.path.exists.assert_called_with(mount_path3)
        self.assertEqual(
            [{'gathered_at': 'fake_date',
              'id': 'fakeid_success', 'used_size': '1'}],
            update_shares)
        self._driver._execute.assert_called_with(
            'df', '-l', '--output=target,used',
            '--block-size=g')

    def test_update_share_usage_size_fail(self):
        def _fake_exec(*args, **kwargs):
            raise exception.ProcessExecutionError(stderr="error")

        self.mock_object(self._driver, '_execute', _fake_exec)
        self.assertRaises(exception.ProcessExecutionError,
                          self._driver.update_share_usage_size,
                          self._context,
                          [self.share])

    def test_get_backend_info(self):
        backend_info = self._driver.get_backend_info(self._context)

        self.assertEqual(
            {'export_ips': ','.join(self.server['public_addresses']),
             'db_version': mock.ANY},
            backend_info)

    def test_get_mount_point_name_with_mount_point_name(self):
        share = {'mount_point_name': 'fake_mp_name', 'name': 'fakename'}
        result = self._driver._get_mount_point_name(share)
        self.assertEqual(result, 'fake_mp_name')

    def test_get_mount_point_name_without_mount_point_name(self):
        share = {'name': 'fakename'}
        result = self._driver._get_mount_point_name(share)
        self.assertEqual(result, 'fakename')

    def test_get_mount_point_name_with_empty_mount_point_name(self):
        share = {'mount_point_name': '', 'name': 'fakename'}
        result = self._driver._get_mount_point_name(share)
        self.assertEqual(result, 'fakename')

    def test_get_mount_point_name_with_none_mount_point_name(self):
        share = {'mount_point_name': None, 'name': 'fakename'}
        result = self._driver._get_mount_point_name(share)
        self.assertEqual(result, 'fakename')

    def test_get_mount_point_name_without_name(self):
        share = {'mount_point_name': 'fake_mp_name'}
        result = self._driver._get_mount_point_name(share)
        self.assertEqual(result, 'fake_mp_name')



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import datetime
import iso8601

from manila.message import message_field
from manila.message import message_levels
from manila.tests.api import fakes


FAKE_UUID = fakes.FAKE_UUID


def stub_message(id, **kwargs):
    message = {
        'id': id,
        'project_id': 'fake_project',
        'action_id': message_field.Action.ALLOCATE_HOST[0],
        'message_level': message_levels.ERROR,
        'request_id': FAKE_UUID,
        'resource_type': message_field.Resource.SHARE,
        'resource_id': 'fake_uuid',
        'updated_at': datetime.datetime(1900, 1, 1, 1, 1, 1,
                                        tzinfo=iso8601.UTC),
        'created_at': datetime.datetime(1900, 1, 1, 1, 1, 1,
                                        tzinfo=iso8601.UTC),
        'expires_at': datetime.datetime(1900, 1, 1, 1, 1, 1,
                                        tzinfo=iso8601.UTC),
        'detail_id': message_field.Detail.NO_VALID_HOST[0],
    }

    message.update(kwargs)
    return message


def stub_message_get(self, context, message_id):
    return stub_message(message_id)


def stub_lock(id, **kwargs):
    lock = {
        'id': id,
        'project_id': 'f63f7a159f404cfc8604b7065c609691',
        'user_id': 'e78f4294e3534e00ae176bd989d6a682',
        'resource_id': 'c474badd-f06e-4ff9-ae26-daa00e19867b',
        'resource_action': 'delete',
        'resource_type': 'share',
        'lock_context': 'user',
        'lock_reason': 'for the tests',
        'updated_at': datetime.datetime(2023, 8, 10, 20, 4, 39,
                                        tzinfo=iso8601.UTC),
        'created_at': datetime.datetime(2023, 1, 10, 15, 3, 1,
                                        tzinfo=iso8601.UTC),
    }

    lock.update(kwargs)
    return lock



# Copyright (c) 2022 China Telecom Digital Intelligence.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import http.client as http_client
from unittest import mock

import ddt
from oslo_serialization import jsonutils
import webob

from manila.api.v2 import share_transfer
from manila import context
from manila import db
from manila import exception
from manila import quota
from manila.share import api as share_api
from manila.share import rpcapi as share_rpcapi
from manila.share import share_types
from manila import test
from manila.tests.api import fakes
from manila.tests import db_utils
from manila.transfer import api as transfer_api

SHARE_TRANSFER_VERSION = "2.77"


@ddt.ddt
class ShareTransferAPITestCase(test.TestCase):
    """Test Case for transfers V3 API."""

    microversion = SHARE_TRANSFER_VERSION

    def setUp(self):
        super(ShareTransferAPITestCase, self).setUp()
        self.share_transfer_api = transfer_api.API()
        self.v2_controller = share_transfer.ShareTransferController()
        self.ctxt = context.RequestContext(
            'fake_user_id', 'fake_project_id', auth_token=True, is_admin=True)

    def _create_transfer(self, share_id='fake_share_id',
                         display_name='test_transfer'):
        transfer = self.share_transfer_api.create(context.get_admin_context(),
                                                  share_id, display_name)
        return transfer

    def _create_share(self, display_name='test_share',
                      display_description='this is a test share',
                      status='available',
                      size=1,
                      project_id='fake_project_id',
                      user_id='fake_user_id',
                      share_network_id=None,
                      mount_point_name=None):
        """Create a share object."""
        share_type = db_utils.create_share_type()
        instance_list = []
        if mount_point_name:
            instance_list.append(
                db_utils.create_share_instance(
                    status=status,
                    share_id='fake_id',
                    mount_point_name=mount_point_name
                )
            )

        share = db_utils.create_share(
            display_name=display_name,
            display_description=display_description,
            status=status,
            size=size,
            project_id=project_id,
            user_id=user_id,
            share_type_id=share_type['id'],
            share_network_id=share_network_id,
            instances=instance_list,
            mount_point_name=mount_point_name
        )

        share_id = share['id']

        return share_id

    def test_show_transfer(self):
        share_id = self._create_share(size=5)
        transfer = self._create_transfer(share_id)
        path = '/v2/fake_project_id/share-transfers/%s' % transfer['id']
        req = fakes.HTTPRequest.blank(path, version=self.microversion)
        req.environ['manila.context'] = self.ctxt
        req.method = 'GET'
        req.headers['Content-Type'] = 'application/json'
        res_dict = self.v2_controller.show(req, transfer['id'])

        self.assertEqual('test_transfer', res_dict['transfer']['name'])
        self.assertEqual(transfer['id'], res_dict['transfer']['id'])
        self.assertEqual(share_id, res_dict['transfer']['resource_id'])

    def test_list_transfers(self):
        share_id_1 = self._create_share(size=5)
        share_id_2 = self._create_share(size=5)
        transfer1 = self._create_transfer(share_id_1)
        transfer2 = self._create_transfer(share_id_2)

        path = '/v2/fake_project_id/share-transfers'
        req = fakes.HTTPRequest.blank(path, version=self.microversion)
        req.environ['manila.context'] = self.ctxt
        req.method = 'GET'
        req.headers['Content-Type'] = 'application/json'
        res_dict = self.v2_controller.index(req)

        self.assertEqual(transfer1['id'], res_dict['transfers'][1]['id'])
        self.assertEqual('test_transfer', res_dict['transfers'][1]['name'])
        self.assertEqual(transfer2['id'], res_dict['transfers'][0]['id'])
        self.assertEqual('test_transfer', res_dict['transfers'][0]['name'])

    def test_list_transfers_with_all_tenants(self):
        share_id_1 = self._create_share(size=5)
        share_id_2 = self._create_share(size=5, project_id='fake_project_id2',
                                        user_id='fake_user_id2')
        self._create_transfer(share_id_1)
        self._create_transfer(share_id_2)

        path = '/v2/fake_project_id/share-transfers?all_tenants=true'
        req = fakes.HTTPRequest.blank(path, version=self.microversion)
        req.environ['manila.context'] = context.get_admin_context()
        req.method = 'GET'
        req.headers['Content-Type'] = 'application/json'
        res_dict = self.v2_controller.index(req)

        self.assertEqual(2, len(res_dict['transfers']))

    def test_list_transfers_with_limit(self):
        share_id_1 = self._create_share(size=5)
        share_id_2 = self._create_share(size=5)
        self._create_transfer(share_id_1)
        self._create_transfer(share_id_2)
        path = '/v2/fake_project_id/share-transfers?limit=1'
        req = fakes.HTTPRequest.blank(path, version=self.microversion)
        req.environ['manila.context'] = self.ctxt
        req.method = 'GET'
        req.headers['Content-Type'] = 'application/json'
        res_dict = self.v2_controller.index(req)

        self.assertEqual(1, len(res_dict['transfers']))

    @ddt.data("desc", "asc")
    def test_list_transfers_with_sort(self, sort_dir):
        share_id_1 = self._create_share(size=5)
        share_id_2 = self._create_share(size=5)
        transfer1 = self._create_transfer(share_id_1)
        transfer2 = self._create_transfer(share_id_2)
        path = \
            '/v2/fake_project_id/share-transfers?sort_key=id&sort_dir=%s' % (
                sort_dir)
        req = fakes.HTTPRequest.blank(path, version=self.microversion)
        req.environ['manila.context'] = self.ctxt
        req.method = 'GET'
        req.headers['Content-Type'] = 'application/json'
        res_dict = self.v2_controller.index(req)

        self.assertEqual(2, len(res_dict['transfers']))
        order_ids = sorted([transfer1['id'],
                            transfer2['id']])
        expect_result = order_ids[1] if sort_dir == "desc" else order_ids[0]
        self.assertEqual(expect_result,
                         res_dict['transfers'][0]['id'])

    def test_list_transfers_detail(self):
        share_id_1 = self._create_share(size=5)
        share_id_2 = self._create_share(size=5)
        transfer1 = self._create_transfer(share_id_1)
        transfer2 = self._create_transfer(share_id_2)

        path = '/v2/fake_project_id/share-transfers/detail'
        req = fakes.HTTPRequest.blank(path, version=self.microversion)
        req.environ['manila.context'] = self.ctxt
        req.method = 'GET'
        req.headers['Content-Type'] = 'application/json'
        req.headers['Accept'] = 'application/json'
        res_dict = self.v2_controller.detail(req)

        self.assertEqual('test_transfer',
                         res_dict['transfers'][1]['name'])
        self.assertEqual(transfer1['id'], res_dict['transfers'][1]['id'])
        self.assertEqual(share_id_1, res_dict['transfers'][1]['resource_id'])

        self.assertEqual('test_transfer',
                         res_dict['transfers'][0]['name'])
        self.assertEqual(transfer2['id'], res_dict['transfers'][0]['id'])
        self.assertEqual(share_id_2, res_dict['transfers'][0]['resource_id'])

    def test_create_transfer(self):
        share_id = self._create_share(status='available', size=5)
        body = {"transfer": {"name": "transfer1",
                             "share_id": share_id}}

        path = '/v2/fake_project_id/share-transfers'
        req = fakes.HTTPRequest.blank(path, version=self.microversion)
        req.environ['manila.context'] = self.ctxt
        req.method = 'POST'
        req.headers['Content-Type'] = 'application/json'
        req.body = jsonutils.dumps(body).encode("utf-8")
        res_dict = self.v2_controller.create(req, body)

        self.assertIn('id', res_dict['transfer'])
        self.assertIn('auth_key', res_dict['transfer'])
        self.assertIn('created_at', res_dict['transfer'])
        self.assertIn('name', res_dict['transfer'])
        self.assertIn('resource_id', res_dict['transfer'])

    @ddt.data({},
              {"transfer": {"name": "transfer1"}},
              {"transfer": {"name": "transfer1",
                            "share_id": "invalid_share_id"}})
    def test_create_transfer_with_invalid_body(self, body):
        path = '/v2/fake_project_id/share-transfers'
        req = fakes.HTTPRequest.blank(path, version=self.microversion)
        req.environ['manila.context'] = self.ctxt
        req.method = 'POST'
        req.headers['Content-Type'] = 'application/json'
        req.body = jsonutils.dumps(body).encode("utf-8")
        self.assertRaises(webob.exc.HTTPBadRequest,
                          self.v2_controller.create, req, body)

    def test_create_transfer_with_invalid_share_status(self):
        share_id = self._create_share()
        body = {"transfer": {"name": "transfer1",
                             "share_id": share_id}}
        db.share_update(context.get_admin_context(),
                        share_id, {'status': 'error'})

        path = '/v2/fake_project_id/share-transfers'
        req = fakes.HTTPRequest.blank(path, version=self.microversion)
        req.environ['manila.context'] = self.ctxt
        req.method = 'POST'
        req.headers['Content-Type'] = 'application/json'
        req.body = jsonutils.dumps(body).encode("utf-8")
        self.assertRaises(webob.exc.HTTPBadRequest,
                          self.v2_controller.create, req, body)

    def test_create_transfer_with_invalid_mount_point_name(self):
        share_id = self._create_share(
            project_id='fake_pid',
            mount_point_name='fake_pid_mount_point_name')
        body = {"transfer": {"name": "transfer1",
                             "share_id": share_id}}
        db.share_update(context.get_admin_context(),
                        share_id, {'status': 'error'})

        path = '/v2/fake_project_id/share-transfers'
        req = fakes.HTTPRequest.blank(path, version=self.microversion)
        req.environ['manila.context'] = self.ctxt
        req.method = 'POST'
        req.headers['Content-Type'] = 'application/json'
        req.body = jsonutils.dumps(body).encode("utf-8")
        self.assertRaises(webob.exc.HTTPBadRequest,
                          self.v2_controller.create, req, body)

    def test_create_transfer_with_project_id_prefix_mount_point_name(self):
        share_id = self._create_share(project_id='fake',
                                      mount_point_name='fake_mp')

        self.assertRaises(exception.Invalid,
                          self.share_transfer_api.create,
                          context.get_admin_context(), share_id,
                          'test_missing_share_type')

    def test_create_transfer_share_with_network_id(self):
        share_id = self._create_share(share_network_id='fake_id')
        body = {"transfer": {"name": "transfer1",
                             "share_id": share_id}}

        path = '/v2/fake_project_id/share-transfers'
        req = fakes.HTTPRequest.blank(path, version=self.microversion)
        req.environ['manila.context'] = self.ctxt
        req.method = 'POST'
        req.headers['Content-Type'] = 'application/json'
        req.body = jsonutils.dumps(body).encode("utf-8")
        self.assertRaises(webob.exc.HTTPBadRequest,
                          self.v2_controller.create, req, body)

    def test_create_transfer_share_with_invalid_snapshot(self):
        share_id = self._create_share(share_network_id='fake_id')
        db_utils.create_snapshot(share_id=share_id)
        body = {"transfer": {"name": "transfer1",
                             "share_id": share_id}}

        path = '/v2/fake_project_id/share-transfers'
        req = fakes.HTTPRequest.blank(path, version=self.microversion)
        req.environ['manila.context'] = self.ctxt
        req.method = 'POST'
        req.headers['Content-Type'] = 'application/json'
        req.body = jsonutils.dumps(body).encode("utf-8")
        self.assertRaises(webob.exc.HTTPBadRequest,
                          self.v2_controller.create, req, body)

    def test_delete_transfer_awaiting_transfer(self):
        share_id = self._create_share()
        transfer = self.share_transfer_api.create(context.get_admin_context(),
                                                  share_id, 'test_transfer')
        path = '/v2/fake_project_id/share-transfers/%s' % transfer['id']
        req = fakes.HTTPRequest.blank(path, version=self.microversion)
        req.environ['manila.context'] = self.ctxt
        req.method = 'DELETE'
        req.headers['Content-Type'] = 'application/json'
        self.v2_controller.delete(req, transfer['id'])

        # verify transfer has been deleted
        req = fakes.HTTPRequest.blank(path, version=self.microversion)
        req.environ['manila.context'] = self.ctxt
        req.method = 'GET'
        req.headers['Content-Type'] = 'application/json'
        res = req.get_response(fakes.app())

        self.assertEqual(http_client.NOT_FOUND, res.status_int)
        self.assertEqual(db.share_get(context.get_admin_context(),
                         share_id)['status'], 'available')

    def test_delete_transfer_not_awaiting_transfer(self):
        share_id = self._create_share()
        transfer = self.share_transfer_api.create(context.get_admin_context(),
                                                  share_id, 'test_transfer')
        db.share_update(context.get_admin_context(),
                        share_id, {'status': 'available'})

        path = '/v2/fake_project_id/share-transfers/%s' % transfer['id']
        req = fakes.HTTPRequest.blank(path, version=self.microversion)
        req.environ['manila.context'] = self.ctxt
        req.method = 'DELETE'
        req.headers['Content-Type'] = 'application/json'
        self.assertRaises(exception.InvalidShare,
                          self.v2_controller.delete, req,
                          transfer['id'])

    def test_transfer_accept_share_id_specified(self):
        share_id = self._create_share()
        transfer = self.share_transfer_api.create(context.get_admin_context(),
                                                  share_id, 'test_transfer')
        self.mock_object(quota.QUOTAS, 'reserve', mock.Mock())
        self.mock_object(quota.QUOTAS, 'commit', mock.Mock())
        self.mock_object(share_api.API,
                         'check_is_share_size_within_per_share_quota_limit',
                         mock.Mock())
        self.mock_object(share_rpcapi.ShareAPI,
                         'transfer_accept',
                         mock.Mock())
        fake_share_type = {'id': 'fake_id',
                           'name': 'fake_name',
                           'is_public': True}
        self.mock_object(share_types, 'get_share_type',
                         mock.Mock(return_value=fake_share_type))
        self.mock_object(db, 'share_snapshot_get_all_for_share',
                         mock.Mock(return_value={}))

        body = {"accept": {"auth_key": transfer['auth_key']}}
        path = '/v2/fake_project_id/share-transfers/%s/accept' % transfer['id']
        req = fakes.HTTPRequest.blank(path, version=self.microversion)
        req.environ['manila.context'] = self.ctxt
        req.method = 'POST'
        req.headers['Content-Type'] = 'application/json'
        req.body = jsonutils.dumps(body).encode("utf-8")
        self.v2_controller.accept(req, transfer['id'], body)

    def test_transfer_accept_with_not_public_share_type(self):
        share_id = self._create_share()
        transfer = self.share_transfer_api.create(context.get_admin_context(),
                                                  share_id, 'test_transfer')
        fake_share_type = {'id': 'fake_id',
                           'name': 'fake_name',
                           'is_public': False,
                           'projects': ['project_id1', 'project_id2']}
        self.mock_object(share_types, 'get_share_type',
                         mock.Mock(return_value=fake_share_type))

        body = {"accept": {"auth_key": transfer['auth_key']}}
        path = '/v2/fake_project_id/share-transfers/%s/accept' % transfer['id']
        req = fakes.HTTPRequest.blank(path, version=self.microversion)
        req.environ['manila.context'] = self.ctxt
        req.method = 'POST'
        req.headers['Content-Type'] = 'application/json'
        req.body = jsonutils.dumps(body).encode("utf-8")
        self.assertRaises(webob.exc.HTTPBadRequest,
                          self.v2_controller.accept, req,
                          transfer['id'], body)

    @ddt.data({},
              {"accept": {}},
              {"accept": {"auth_key": "fake_auth_key",
                          "clear_access_rules": "invalid_bool"}})
    def test_transfer_accept_with_invalid_body(self, body):
        path = '/v2/fake_project_id/share-transfers/fake_transfer_id/accept'
        req = fakes.HTTPRequest.blank(path, version=self.microversion)
        req.environ['manila.context'] = self.ctxt
        req.method = 'POST'
        req.headers['Content-Type'] = 'application/json'
        req.body = jsonutils.dumps(body).encode("utf-8")
        self.assertRaises(webob.exc.HTTPBadRequest,
                          self.v2_controller.accept, req,
                          'fake_transfer_id', body)

    def test_transfer_accept_with_invalid_auth_key(self):
        share_id = self._create_share(size=5)
        transfer = self._create_transfer(share_id)
        body = {"accept": {"auth_key": "invalid_auth_key"}}
        path = '/v2/fake_project_id/share-transfers/%s/accept' % transfer['id']
        req = fakes.HTTPRequest.blank(path, version=self.microversion)
        req.environ['manila.context'] = self.ctxt
        req.method = 'POST'
        req.headers['Content-Type'] = 'application/json'
        req.body = jsonutils.dumps(body).encode("utf-8")
        self.assertRaises(webob.exc.HTTPBadRequest,
                          self.v2_controller.accept, req,
                          transfer['id'], body)

    def test_transfer_accept_with_invalid_share_status(self):
        share_id = self._create_share(size=5)
        transfer = self._create_transfer(share_id)
        db.share_update(context.get_admin_context(),
                        share_id, {'status': 'error'})
        body = {"accept": {"auth_key": transfer['auth_key']}}
        path = '/v2/fake_project_id/share-transfers/%s/accept' % transfer['id']
        req = fakes.HTTPRequest.blank(path, version=self.microversion)
        req.environ['manila.context'] = self.ctxt
        req.method = 'POST'
        req.headers['Content-Type'] = 'application/json'
        req.body = jsonutils.dumps(body).encode("utf-8")
        self.assertRaises(webob.exc.HTTPBadRequest,
                          self.v2_controller.accept, req,
                          transfer['id'], body)

    @ddt.data({'overs': {'gigabytes': 'fake'}},
              {'overs': {'shares': 'fake'}},
              {'overs': {'snapshot_gigabytes': 'fake'}},
              {'overs': {'snapshots': 'fake'}})
    @ddt.unpack
    def test_accept_share_over_quota(self, overs):
        share_id = self._create_share()
        db_utils.create_snapshot(share_id=share_id, status='available')
        transfer = self.share_transfer_api.create(context.get_admin_context(),
                                                  share_id, 'test_transfer')

        usages = {'gigabytes': {'reserved': 5, 'in_use': 5},
                  'shares': {'reserved': 10, 'in_use': 10},
                  'snapshot_gigabytes': {'reserved': 5, 'in_use': 5},
                  'snapshots': {'reserved': 10, 'in_use': 10}}

        quotas = {'gigabytes': 5, 'shares': 10,
                  'snapshot_gigabytes': 5, 'snapshots': 10}
        exc = exception.OverQuota(overs=overs, usages=usages, quotas=quotas)
        self.mock_object(quota.QUOTAS, 'reserve', mock.Mock(side_effect=exc))
        self.mock_object(quota.QUOTAS, 'commit', mock.Mock())
        self.mock_object(share_api.API,
                         'check_is_share_size_within_per_share_quota_limit',
                         mock.Mock())
        self.mock_object(share_rpcapi.ShareAPI,
                         'transfer_accept',
                         mock.Mock())
        fake_share_type = {'id': 'fake_id',
                           'name': 'fake_name',
                           'is_public': True}
        self.mock_object(share_types, 'get_share_type',
                         mock.Mock(return_value=fake_share_type))

        body = {"accept": {"auth_key": transfer['auth_key']}}
        path = '/v2/fake_project_id/share-transfers/%s/accept' % transfer['id']
        req = fakes.HTTPRequest.blank(path, version=self.microversion)
        req.environ['manila.context'] = self.ctxt
        req.method = 'POST'
        req.headers['Content-Type'] = 'application/json'
        req.body = jsonutils.dumps(body).encode("utf-8")
        self.assertRaises(webob.exc.HTTPRequestEntityTooLarge,
                          self.v2_controller.accept, req,
                          transfer['id'], body)
